import json
import inspect
import logging
import datetime
import pandas

from .estimator import Estimator

class DataProvider(object):

    ###########################################################
    ###
    ###   Attributes every DataProvider must define
    ###
    ###########################################################

    id:                            str    = None
    x_features:                    list   = None
    x_estimator_features:          list
    y:                             str
    train_dataset_sources:         dict   = dict()
    batch_predict_dataset_sources: dict   = dict()
    hollow:                        bool   = False



    ###########################################################
    ###
    ###   Attributes for data splitting and randomization
    ###
    ###########################################################

    random_state:                  int    = 42
    test_size:                     float  = 0.1
    val_size:                      float  = 0.2

    ###########################################################################
    ###
    ###   Attributes for the Estimator
    ###
    ###   See methods:
    ###     get_estimator_class()
    ###     get_estimator_optimization_search_space()
    ###     get_estimator_parameters()
    ###     get_estimator_hyperparameters()
    ###
    # The (hyper)params is what a SciKit estimator gets in its
    # __init__() method. They are of 2 types:
    #
    #  - estimation quality parameters (those that are tunned and are commonly
    #      called hyper-parameters)
    #  - operational parameters (control operation, verbosity etc)
    #
    # estimator_hyperparams + estimator_params is what is used to initialize
    # the object. Like:
    #     XGBClassifier(**estimator_params,**estimator_hyperparams)
    #
    # We are separating them here because estimator_hyperparams will be
    # optimized while estimator_params are kind of fixed.
    #
    # estimator_hyperparams_search_space and estimator_hyperparams usually
    # contain same keys. The search_space has value ranges and semantics
    # determined by your optimization framework as optuna or skopt.
    #
    # The estimator_class_params is random parameters for the xingu.Estimator
    # class.
    #
    # A xingu.Estimator class will be initialized by xingu.Model like this:
    #
    #    xingu.Estimator(
    #         **DataProvider.estimator_class_params,
    #         params      = DataProvider.estimator_params,
    #         hyperparams = DataProvider.estimator_hyperparams,
    #    )
    #
    ###########################################################################

    estimator_class:                      type   = Estimator
    estimator_class_params:               dict   = dict()
    estimator_params:                     dict   = dict()
    estimator_hyperparams:                dict   = dict()
    estimator_hyperparams_search_space:   dict   = dict()


    ###########################################################
    ###
    ###   Control batch predict and metrics computers.
    ###
    ###########################################################

    # For classification models, the index of predict_proba result tha matches
    # the training target
    proba_class_index:                    int    = 0


    def __init__(self):
        if type(self.estimator_class) == str:
            # Convert an estimator class string into its real class.
            # For example, the string 'xingu.estimators.xgboost_optuna.XinguXGBoostClassifier'
            # will be converted to its real class.
            import importlib

            mod=importlib.import_module('.'.join(self.estimator_class.split('.')[:-1]))
            self.estimator_class=getattr(mod,self.estimator_class.split('.')[-1])

        self.log(self.estimator_class)
        if issubclass(self.estimator_class, Estimator)==False:
            raise RuntimeError(f"Invalid class: {self.estimator_class}. Must be subclass of xingu.Estimator.")


    ###########################################################
    ###
    ###   Methods that need specialized implementation on
    ###   derived Data Providers.
    ###
    ###########################################################


    def clean_data_for_train(self, datasets: dict) -> pandas.DataFrame:
        """
        datasets is a dict with format:

            {
                'NAME_OF_DATASET1': pandas.DataFrame,
                'NAME_OF_DATASET2': pandas.DataFrame,
                ...
            }

        This method needs to integrate all these DataFrames and return a single
        DataFrame with minimum cleanup.

        Provides a default implementation that simply concatenates all
        datasets. This is probably very wrong for your DataProvider, so please
        reaimplement.
        """
        result = None

        for d in datasets:
            if result is None:
                result = datasets[d]
            else:
                result = pandas.concat(
                    [
                        result,
                        datasets[d]
                    ]
                )

        return result



    def feature_engineering_for_train(self, data: pandas.DataFrame) -> pandas.DataFrame:
        """
        Applies any post-cleanup feature engineering on the DataFrame returned
        by clean_data().

        This must be implemented on a derived class.
        """
        return data



    def last_pre_process_for_train(self, data: pandas.DataFrame) -> pandas.DataFrame:
        """
        Last chance to process data returned by train_feature_engineering() and before training.

        This must be implemented on a derived class.
        """
        return data



    def clean_data_for_batch_predict(self, datasets: dict) -> pandas.DataFrame:
        # Default implementation; please reimplement
        return self.clean_data_for_train(datasets)



    def feature_engineering_for_batch_predict(self, data: pandas.DataFrame) -> pandas.DataFrame:
        return data



    def last_pre_process_for_batch_predict(self, data: pandas.DataFrame) -> pandas.DataFrame:
        """
        Last chance to process data returned by feature_engineering_for_batch_predict() and before batch prediction.

        This must be implemented on a derived class.
        """
        return data



    # Implement this if you need to fabricate features for your estimator. And in this
    # case, x_estimator_features and x_features might probably be different
    def feature_engineering_for_predict(self, data: pandas.DataFrame) -> pandas.DataFrame:
        """
        User must call this before all pred_*() methods in order to prepare data for the estimator.

        Here might rest code shared by feature_engineering_for_train() and feature_engineering_for_batch_predict()
        """
        return data



    def get_estimator_class(self) -> type:
        return self.estimator_class



    def get_estimator_class_params(self) -> dict:
        return self.estimator_class_params



    def get_estimator_params(self) -> dict:
        return self.estimator_params



    def get_estimator_hyperparams(self) -> dict:
        return self.estimator_hyperparams



    def get_estimator_optimization_search_space(self) -> dict:
        return self.estimator_hyperparams_search_space



    def post_process_after_hyperparam_optimize(self, model):
        """
        Called after estimator was optimized by hyperparam_optimize() but
        before the final train.

        model.estimator has the xingu.Estimator object.

        The `model` parameter has the entire Model object and the `dp` attribute
        inside `model` equals to `self`. In other words, `model.dp` and `self`
        are the same object in the context of this method. Yet in another set
        of words, `model.dp == self`.
        """
        pass



    def post_process_after_train(self, model):
        """
        Called after estimator was trained but before PKL is saved or metrics computed.
        Implement it in your DataProvider if you need to compute coefficients or
        manipulate your Model object somehow.

        If you’ll compute things that will be used in future predicts, don’t forget to
        save them in the PKL implementing your own __getstate__() method.

        The `model` parameter has the entire Model object and the `dp` attribute
        inside `model` equals to `self`. In other words, `model.dp` and `self` are the
        same object in the context of this method. Yet in another set of
        words, `model.dp == self`.
        """
        pass



    def pre_process_for_generic_predict(self, X: pandas.DataFrame, model) -> pandas.DataFrame:
        """
        Virtual method to be implemented in concrete DPs in case your pre-processing is
        the same for predict() and predict_proba().

        If pre-processing is not the same for these methods, implement both
        pre_process_for_predict() and pre_process_for_predict_proba()
        """
        return X



    def post_process_after_generic_predict(self, X: pandas.DataFrame, Y_pred: pandas.DataFrame, model) -> pandas.DataFrame:
        """
        Virtual method to be implemented in concrete DPs in case your post-processing is
        the same for predict() and predict_proba().

        If post-processing is not the same for these methods, implement both
        post_process_after_predict() and post_process_after_predict_proba()
        """
        return Y_pred



    def pre_process_for_predict(self, X: pandas.DataFrame, model) -> pandas.DataFrame:
        """
        Called by Model.generic_predict() right before X is passed to its internal
        estimator to compute Ŷ.

        Your implementation may modify X completely, and what you return here
        will be passed to Model's internal estimator.

        The abstract implementation here does nothing.
        """
        return self.pre_process_for_generic_predict(X,model)



    def post_process_after_predict(self, X: pandas.DataFrame, Y_pred: pandas.DataFrame, model) -> pandas.DataFrame:
        """
        Called by Model.pred_dist() right after Ŷ (Y_pred) is computed.

        X is whatever you returned in pre_process_for_pred_dist().

        Your implementation may modify Y_pred completely, and what you return here
        will be returned to Model caller.

        The abstract implementation here does nothing.
        """
        return self.post_process_after_generic_predict(X,Y_pred,model)



    def pre_process_for_predict_proba(self, X: pandas.DataFrame, model) -> pandas.DataFrame:
        """
        Called by Model.generic_predict() right before X is passed to its internal
        estimator to compute Ŷ.

        Your implementation may modify X completely, and what you return here
        will be passed to Model's internal estimator.

        The abstract implementation here does nothing.
        """
        return self.pre_process_for_generic_predict(X,model)



    def post_process_after_predict_proba(self, X: pandas.DataFrame, Y_pred: pandas.DataFrame, model) -> pandas.DataFrame:
        """
        Called by Model.pred_dist() right after Ŷ (Y_pred) is computed.

        X is whatever you returned in pre_process_for_pred_dist().

        Your implementation may modify Y_pred completely, and what you return here
        will be returned to Model caller.

        The abstract implementation here does nothing.
        """
        return self.post_process_after_generic_predict(X,Y_pred,model)


    def pre_process_for_trainsets_metrics(self, model):
        """
        Called by Model.trainsets_predict() after train sets where predicted
        but before metrics computation.

        model.sets and model.sets_estimations are dicts with interesting
        DataFrames.
        """
        pass



    def post_process_after_trainsets_metrics(self, model):
        """
        Called by Model.trainsets_predict() after metrics computation.

        model.sets and model.sets_estimations are dicts with interesting
        DataFrames.
        """
        pass


    ###########################################################
    ###
    ###   Generic methods used as is by all derived classes.
    ###   But derived classes can reimplement a specialization
    ###   of them.
    ###
    ###########################################################


    def get_logger(self):
        if not hasattr(self,'logger'):
            self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        return self.logger



    def log(self, message='', level=logging.INFO):
        self.get_logger().log(
            level,
            'id={id}: {message}'.format(
                id=self.id,
                message=message
            )
        )



    def get_dataset_sources_for_train(self, model=None) -> dict:
        """
        Returns the train_dataset_sources, which is a dicts with format:

            {
                'NAME_OF_DATASET1': {
                    'source': 'units',
                    'query': 'select ...'
                },
                'NAME_OF_DATASET2': {
                    'source': 'xingu',
                    'query': 'select ...'
                },
                'NAME_OF_DATASET3': {
                    'source': 'units',
                    'query': 'select ...'
                }
            }
        """
        return self.train_dataset_sources



    def get_dataset_sources_for_batch_predict(self, model=None) -> dict:
        return self.batch_predict_dataset_sources



    def get_features_list(self) -> list:
        return self.x_features



    def get_estimator_features_list(self) -> list:
        if hasattr(self,'x_estimator_features') and len(self.x_estimator_features)>0:
            # If we actually have x_estimator_features...
            return self.x_estimator_features
        else:
            return self.x_features



    def get_target(self) -> str:
        return self.y



    def __repr__(self) -> str:
        return '{klass}(id={id})'.format(id=self.id, klass=type(self).__name__)



    ###########################################################
    ###
    ###   Generic data split methods
    ###
    ###########################################################


    def random_data_split(data: pandas.DataFrame, test_size: float, val_size: float) -> (pandas.DataFrame, pandas.DataFrame, pandas.DataFrame):
        """
        Split data in 3 parts as:

        - train (size = total - (test_size + val_size))
        - val (size = val_size)
        - test (size = test_size)

        Data are split with sklearn.model_selection.train_test_split().

        test_size and val_size must be float, between 0.0 and 1.0 and represent the
        proportion of split.

        Derived classes might have more specialized implementations to split data.
        """

        from sklearn.model_selection import train_test_split

        df_train, df_test = train_test_split(
            data,
            test_size=test_size,
            random_state=self.random_state
        )

        df_train, df_val = train_test_split(
            df_train,
            test_size=val_size,
            random_state=self.random_state
        )

        return (df_train, df_val, df_test)



    def time_data_split(df, date_column, delta_time_days):
        """
        Split data in 2 parts as:

        - train (all data where date_column is prior than delta_time_days ago)
        - test (all data where date_column is after delta_time_days ago)

        In summary, test will have the latest delta_time_days data and train will have all
        the rest, which is older.

        delta_time_days is an integer number of days.

        date_column is the name of date column in the df dataset.
        """
        split_date = datetime.datetime.now() - datetime.timedelta(delta_time_days)

        df_train = df[df[date_column] < split_date.date()]
        df_test = df[df[date_column] >= split_date.date()]

        return df_train, df_test



    def data_split_for_train(self, data: pandas.DataFrame) -> dict:
        """
        Split data before train.
        Implement in your DataProvider.

        Must return a dict of dataframes. The key named 'train' will be used
        for training. Other dataframes can be used for metrics computation or
        other purposes in your DataFrame.
        """
        return dict(train=data)


    ####################################################################
    ###
    ###   Hooks for customized and extensible metrics computation
    ###
    ####################################################################


    def get_metrics_computers(self, type: str='estimation') -> list:
        """
        Return a list of member functions that can compute metrics for estimator outputs.

        Current types that will be scanned are:

        • trainsets_model
        • batch_model
        • global_model
        • estimation
        """


        methods=[]

        for method in inspect.getmembers(self, predicate=inspect.ismethod):
            if 'compute_' + type + '_metrics_' in method[0]:
                methods.append(method[1])

        return methods



    def get_plot_renderers(self, type: str='global_model') -> list:
        """
        Return a list of member functions that can compute metrics for estimator outputs.

        Current types that will be scanned are:

        • trainsets_model
        • batch_model
        • global_model
        """


        methods=[]

        for method in inspect.getmembers(self, predicate=inspect.ismethod):
            if 'render_' + type + '_plots_' in method[0]:
                methods.append(method[1])

        return methods



    def __getstate__(self):
        return dict(
            x_features                         = self.x_features,
            x_estimator_features               = self.x_estimator_features,
            y                                  = self.y,
            estimator_class                    = self.estimator_class,
            estimator_class_params             = self.estimator_class_params,
            estimator_params                   = self.estimator_params,
            estimator_hyperparams              = self.estimator_hyperparams,
            estimator_hyperparams_search_space = self.estimator_hyperparams_search_space,
            proba_class_index                  = self.proba_class_index,
            train_dataset_sources              = self.train_dataset_sources,
            batch_predict_dataset_sources      = self.batch_predict_dataset_sources,
            random_state                       = self.random_state,
            test_size                          = self.test_size,
            val_size                           = self.val_size,
            hollow                             = self.hollow,
            id                                 = self.id,
        )



class DPSimple(DataProvider):
    """
    DataProvider to be used with Xingu´s simpledp CLI arguments.
    """
    def __init__(self,**kwargs):
        if (
                'SIMPLEDP_ID'      in kwargs and
                'TRAIN_DATASOURCE' in kwargs and
                'TARGET_FEATURE'   in kwargs and
                'ESTIMATOR_CLASS'  in kwargs
            ):
            # The 4 mandatory parameters...
            self.id=kwargs['SIMPLEDP_ID']
            self.train_dataset_sources = {
                f'df{ds:04d}': dict(url=kwargs['TRAIN_DATASOURCE'][ds][0])
                for ds in range(len(kwargs['TRAIN_DATASOURCE']))
            }
            self.y=kwargs['TARGET_FEATURE']
            self.estimator_class = kwargs['ESTIMATOR_CLASS']

            # Other parameters
            if 'ESTIMATOR_FEATURES' in kwargs:
                self.x_estimator_features = [
                    p.strip()
                    for p in kwargs['ESTIMATOR_FEATURES'].split(',')
                ]

            if 'PROBA_CLASS_INDEX' in kwargs:
                self.proba_class_index = kwargs['PROBA_CLASS_INDEX']

            if 'ESTIMATOR_CLASS_PARAMS' in kwargs:
                self.estimator_class_params = json.loads(kwargs['ESTIMATOR_CLASS_PARAMS'])

            if 'ESTIMATOR_PARAMS' in kwargs:
                self.estimator_params = json.loads(kwargs['ESTIMATOR_PARAMS'])

            if 'ESTIMATOR_HYPERPARAMS' in kwargs:
                self.estimator_hyperparams = json.loads(kwargs['ESTIMATOR_HYPERPARAMS'])

            if 'ESTIMATOR_HYPERPARAMS_SEARCH_SPACE' in kwargs:
                self.estimator_hyperparams_search_space = json.loads(kwargs['ESTIMATOR_HYPERPARAMS_SEARCH_SPACE'])

            if 'BASE_CLASS' in kwargs:
                self.add_base_class(kwargs['BASE_CLASS'])

        else:
            raise ValueError("Simple DataProvider training requires at least these parameters: TRAIN_DATASOURCE, TARGET_FEATURE, ESTIMATOR_CLASS, SIMPLEDP_ID")

        super().__init__()



    def add_base_class(self, cls, dp_folder: str=None):
        if dp_folder:
            import sys
            sys.path.insert(1, dp_folder)

        if isinstance(cls, str):
            # If this is a class full name (as 'generic_credit_risk.DPGenericCreditRisk'),
            # convert it into a real class object.
            # If cls is not a string, it is expected to be a class object
            # already.

            import importlib

            # Store the base class as a string for future pickling and
            # unpickling
            self._simpledp_base = cls

            mod=importlib.import_module('.'.join(cls.split('.')[:-1]))
            cls=getattr(mod,cls.split('.')[-1])

        # It is expected that cls is a class derived from xingu.DataProvider,
        # as DPSimple class is. Thus it is safe to simply overwrite base
        # classes.
        self.__class__.__bases__ = (cls, )



    def __getstate__(self):
        return dict(
            **super().__getstate__(),

            # Save these additional internals to the pickle
            _simpledp_base = self._simpledp_base
        )



    def __setstate__(self, unpickled):
        self.__dict__.update(unpickled)
        self.add_base_class(self._simpledp_base)



import os
import logging
import concurrent.futures
import decouple
import numpy
import pandas
import catboost

import xingu


class XinguCatBoostClassifier(xingu.Estimator):
    """
    Multi-CatBoost implementation of an estimator suited for Pan operations.
    """


    hyperparam = dict()



    def __init__(self, bagging_size=1, hyperparams: dict=hyperparam, random_state=42):
        super().__init__(hyperparams)

        self.bagging_size=bagging_size
        self.bagging_members=[]

        self.random_state=random_state



    def __repr__(self):
        template='{klass}(size={size}, random_state={random_state}, members={members})'

        text = template.format(
            klass          = type(self).__name__,
            size           = self.bagging_size,
            random_state   = self.random_state,
            members        = self.bagging_members
        )

        # Remove extra spaces and \n
        text = ' '.join(text.split())

        return text



    def hyperparam_optimize(self, datasets: dict, features: list, target: str, search_space: dict):
        from sklearn.model_selection import KFold
        from sklearn.metrics import roc_auc_score, make_scorer
        from skopt import BayesSearchCV

        optimizer = BayesSearchCV(
            # Semantics
            estimator            = catboost.CatBoostClassifier(**self.hyperparam),
            cv                   = KFold(n_splits=5, shuffle=True, random_state=2022),
            search_spaces        = search_space,
            scoring              = make_scorer(
                roc_auc_score,
                needs_threshold     = True
            ),

            # Verbosity
            verbose              = 1,

            # Operational
            n_iter               = 200,
            # use just 1 job with CatBoost in order to avoid segmentation fault
            # n_jobs               = 1,
            return_train_score   = False,
            refit                = True,
            optimizer_kwargs     = dict(base_estimator='GP'),
            random_state         = self.random_state,
        )

        optimizer.fit(
            X = datasets['train'][features],
            y = datasets['train'][target],

            # callbacks = [skopt.callbacks.VerboseCallback(100)]
        )

        # Convert the OrderedDict returned by these objects into a plain dict
        return {i[0]:i[1] for i in optimizer.best_params_.items()}



    def fit_single(self, datasets, features, target, index) -> catboost.CatBoostClassifier:
        train_bootstrap = datasets['train'].sample(
            frac               = 1,
            replace            = True,
            random_state       = self.random_state+index
        )

        clf = catboost.CatBoostClassifier(
            **self.hyperparam
        )

        # Actual training session begins
        self.log(f'Trigger parallel train of CatBoost estimator #{index+1} of {self.bagging_size}...')

        clf.fit(
            train_bootstrap[features],
            train_bootstrap[target],
        )

        return clf



    def fit(self, datasets, features, target, model=None):
        # Add attribute 'max_workers=1' to inhibit parallelism

        max_workers=decouple.config('PARALLEL_ESTIMATORS_MAX_WORKERS', default=0, cast=int)
        if max_workers == '' or max_workers == 0:
            max_workers=None
            self.logger.info(f'Unlimited parallel estimators to train')
        else:
            self.logger.info(f'{max_workers} parallel estimators to train')

        executor=concurrent.futures.ThreadPoolExecutor(thread_name_prefix='fit', max_workers=max_workers)
        tasks=[]
        for i in range(self.bagging_size):
            tasks.append(
                executor.submit(
                    # Method name to call asynchronously
                    self.fit_single,
                    # Its parameters
                    datasets,
                    features,
                    target,
                    i
                )
            )

        self.log(f'Waiting for all member training to finish in parallel')

        # Process result as soon as it is available
        index=0
        for task in concurrent.futures.as_completed(tasks):
            e=task.exception()
            if e is None:
                # Success
                self.bagging_members.append(task.result())
                index+=1
                self.log(f'Finished train of member #{index} of {self.bagging_size}...')
            else:
                # Failure
                raise e



    def predict(self, data: pandas.DataFrame) -> pandas.DataFrame:
        return self.generic_predict(data)



    def predict_proba(self, data: pandas.DataFrame, class_index: int=None) -> pandas.DataFrame:
        return self.generic_predict(
            data,
            method='predict_proba',
            class_index=class_index
        )



    def generic_predict(self, data: pandas.DataFrame, method='predict', class_index: int=None) -> pandas.DataFrame:
        """
        Predict using predict_multi() and then consolidate and aggregate estimations
        for all bagging members.
        """

        duplicate_index_error_message=[
            'DataFrame index has duplicates, which will lead',
            'to sever inconsistencies in multi-model sumarization.',
            'Here is a sample of problematic IDs; clean',
            'DataFrame appropriatelly: {}'
        ]
        
        duplicate_index_error_message=' '.join(duplicate_index_error_message)
        
        if data.shape[0]<1:
            # Empty input. Return an empty standard output.
            return pandas.DataFrame(columns=['estimation'])

        index_duplicate_count=data.index.value_counts()
        if index_duplicate_count.iloc[0] > 1:
            raise OverflowError(
                duplicate_index_error_message
                .format(list(index_duplicate_count[index_duplicate_count>1].head(10).index))
            )

        def ddebug(table,message):
            self.log(message,level=logging.DEBUG)
            return table
        

        if self.bagging_size>1:
            # Define aggregation strategy and compute aggregations for multiple estimations
            # in an ensamble/bagging environment.
            
            # Default aggregation function is mean().
            # But if this is a classifier, predict() returns a class, so we need to find the mode().
            # Aggregation strategy applies the correct aggregation function based if predict() or predict_proba()
            agg_function='mean'
            if self.is_classifier() and method == 'predict':
                agg_function=pandas.Series.mode

            agg_strategy=dict(estimation=agg_function)
            if method == 'predict_proba' and class_index is None:
                agg_strategy={
                    f'estimation_class_{col}':agg_function
                    for col in self.bagging_members[0].classes_
                }

            self.log(f"Bagging aggregation strategy: {agg_strategy}",level=logging.DEBUG)

            return (
                self.predict_multi(data,method,class_index)
                
                # Group by index
                .groupby(level=0)

                # Member ID was consolidated and ghosted at this point.

                # Aggregated results by mean or mode (only if classifier and predict method).
                # One of the reasons Index can’t have duplicates is to avoid
                # bias in here.
                .agg(agg_strategy)

                # Order DataFrame the same as input
                .reindex(data.index)
            )
        else:
            return self.predict_single(data,method,class_index).drop(columns=['member'])



    ####################################################################################
    ###
    ### Internal methods
    ###
    ####################################################################################



    def predict_single(self, data: pandas.DataFrame, method='predict', class_index: int=None, bagging_member: int=0) -> pandas.DataFrame:
        """
        Return raw estimations of one estimator member tagged by member ID.
        If method is 'predict_proba' all class probabilities are returned in multiple
        columns called estimation_class_{I}. Unless class_index is passed, then only
        that class is returned.
        
        Resulting dataframe has columns:

        - index
        - estimator member ID (0,1,2,3 etc as passed in bagging_member)
        - estimation (if predict) or estimation_class_{I} (if predict_proba)

        This method was designed to be parallelized with concurrent.futures.
        """
        
        self.log(f'Member #{bagging_member} is predicting for {data.shape[0]} datapoints...',level=logging.DEBUG)
        
        def ddebug(table,message):
            self.log(message,level=logging.DEBUG)
            return table
        
        renamer=lambda col: 'estimation'
        unwanted_classes=list()
        
        if method == 'predict_proba':
            if class_index is not None:
                # Want single value, not probabilities of all classes
                unwanted_classes=list(set(self.bagging_members[bagging_member].classes_)-set([class_index]))
            else:
                # Change strategy for column renamer
                renamer=lambda col: f'estimation_class_{col}'
        
        return (
            pandas.DataFrame(
                index=data.index,
                
                # Compute estimation via predict or predict_proba, using 1 member
                data=getattr(self.bagging_members[bagging_member], method)(data)
            )
            
            # Keep only the class_index class, if provided
            .drop(columns=unwanted_classes)
            
            # Rename columns from 0, 1, etc to estimation (if predict) or estimation_class_{I}
            .rename(columns=renamer)
            
            # Tag it with member ID
            .assign(member=bagging_member)
            
            # Reduce RAM usage
            .assign(
                member=lambda table: table.member.astype('category')
            )
            
            .pipe(lambda table: ddebug(table,f'Member #{bagging_member} finished predicting for {data.shape[0]} datapoints.'))
        )



    def predict_multi(self, data: pandas.DataFrame, method='predict', class_index: int=None) -> pandas.DataFrame:
        """
        Return raw estimations of all estimator members tagged by member ID in
        a MultiIndex DataFrame.

        Resulting dataframe has columns:

        Index:
        - (original index, preferably unit_id)
        - estimator member ID (0,1,2,3 etc)

        Columns:
        - estimation (if predict) or estimation_class_{I} (if predict_proba)
        """

        executor=concurrent.futures.ThreadPoolExecutor(thread_name_prefix='pred_multi_dist')
        estimation=None
        tasks=[]

        # Trigger all estimators in parallel
        for i in range(self.bagging_size):
            tasks.append(
                # Execute in the background ...
                executor.submit(
                    # ... this method ...
                    self.predict_single,

                    # ... first parameter ...
                    data,

                    # ... second parameter ...
                    method,

                    # ... third parameter ...
                    class_index,

                    # ... fourth parameter ...
                    i,
                )
            )

        # Process their resulting DataFrame as soon as it is available
        for task in concurrent.futures.as_completed(tasks):
            if estimation is None:
                estimation=task.result()
            else:
                estimation=pandas.concat([estimation,task.result()])

        return estimation.set_index([estimation.index,'member']).sort_index()



    def __getstate__(self):
        return dict(
            **xingu.Estimator.__getstate__(self),

            # Number of trained models. Should be same as len(bagging_members)
            bagging_size     = self.bagging_size,

            # Array of trained sklearn.tree.DecisionTreeRegressor()s
            bagging_members  = self.bagging_members,

            # Random number used by the class, as 42
            random_state     = self.random_state
        )



    def is_classifier(self):
        from sklearn.base import is_classifier
        return is_classifier(self.bagging_members[0])


import os
import logging
import pandas



class Estimator(object):
    """
    Base class to encapsulate an estimator as defined by SciKit-Learn

    - Concrete class might have multiple (20) encapsulated predictors for bagging/ensamble.
    """

    default_params = dict()
    default_hyperparams = dict()

    def __init__(self, params: dict=None, hyperparams: dict=None, **kwargs):
        self.setup_logger()

        self.params=params or self.default_params
        self.hyperparams=hyperparams or self.default_hyperparams



    def hyperparam_optimize(self, datasets: dict, model) -> dict:
        # Please implement in a concrete class
        pass



    def hyperparam_exchange(self, params: dict=None, hyperparams: dict=None) -> dict:
        """
        If None is passed, simply return estimator's current hyperparams.
        Set hyperparams otherwise.
        """

        if hyperparams is not None:
            self.hyperparams=hyperparams

        return self.hyperparams



    def fit(self, datasets: dict, features: list, target: str, model=None):
        """
        Train one or multiple models. Leave everything ready for a predict()
        """
        pass



    def predict(self, data: pandas.DataFrame) -> pandas.DataFrame:
        """
        Return a dataframe with estimated Ŷ and indexed as input DataFrame.
        """
        pass



    def predict_proba(self, data: pandas.DataFrame, class_index: int=None) -> pandas.DataFrame:
        """
        Estimate probability of all classes.
        If class_index is not None, only the class with that index will be returned.

        Return a dataframe with estimated Ŷ and indexed as input DataFrame.
        """
        pass



    def is_classifier(self):
        pass

    

    def __repr__(self):
        return '{klass}()'.format(klass=type(self).__name__)



    def __getstate__(self):
        # Do not serialize the logger
        return dict(
            default_params      = self.default_params,
            default_hyperparams = self.default_hyperparams,
            params              = self.params,
            hyperparams         = self.hyperparams,
        )



    def setup_logger(self):
        if not hasattr(self,'logger'):
            # Setup logging
            self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)



    def log(self, message='', level=logging.INFO):
        self.setup_logger()
        self.logger.log(
            level,
            message
        )





import os
import tempfile
import datetime
import pathlib
import logging
import concurrent.futures
import decouple
import numpy
import pandas
import xgboost
import optuna

import xingu


class XinguXGBoostClassifier(xingu.Estimator):
    """
    Multi-XGBoost implementation of a xingu.Estimator optimized by Optuna.
    """


    def __init__(
                self,
                params: dict=None,
                hyperparams: dict=None,
                random_state=42,
                bagging_size=1,
                optimization_trials=10,
                optimization_timeout=None,
                report_interval=None,
                **kwargs
    ):
        """
        - bagging_size defines the number of estimators to be trained along
          with number of folds to StratifiedKFold(). Dataset must have a
          column named "stratify" which will be used by StratifiedKFold().

        - params is a dict of static parameters to XGBClassifier().

        - hyperparams is a dict of parameters to XGBClassifier(). It will
          be ignored if this Estimator will be optimized. In this case, the
          optimization process will query DataProvider.estimator_hyperparam_search_space
          to use as a search space.

        - random_state will be use everywhere possible including XGBClassifier().

        - optimization_trials and optimization_timeout set the length of
          Optuna's optimization. Optimization will stop when number of trials
          are executed or timeout (seconds) is reached, whatever happens first.

        - report_interval defines interval in seconds to report optimization
          status, some statistics, estimated time to finish, and takes the
          chance to dump Optuna's partial data.

        """
        self.random_state = random_state
        if params is None:
            params = dict()

        params['random_state'] = self.random_state
        super().__init__(params=params, hyperparams=hyperparams)

        self.bagging_size = bagging_size
        self.optimization_trials = optimization_trials
        self.optimization_timeout = optimization_timeout
        self.report_interval = report_interval
        self.bagging_members = []



    def __repr__(self):
        template = '{klass}(size={size}, random_state={random_state}, members={members})'

        text = template.format(
            klass          = type(self).__name__,
            size           = self.bagging_size,
            random_state   = self.random_state,
            members        = self.bagging_members
        )

        # Remove extra spaces and \n
        text = ' '.join(text.split())

        return text



    def hyperparam_optimize(self, model):
        import sklearn

        def objective(trial, model):
            """
            Your DataProvider must have something like:

                estimator_hyperparam_search_space = dict(
                    iterations              = ('int',        dict(low=10,    high=500)),
                    depth                   = ('int',        dict(low=1,     high=9)),
                    border_count            = ('int',        dict(low=1,     high=255)),
                    l2_leaf_reg             = ('int',        dict(low=2,     high=35)),
                    learning_rate           = ('loguniform', dict(low=0.01,  high=1.0)),
                    random_strength         = ('loguniform', dict(low=1e-9,  high=10)),
                    bagging_temperature     = ('float',      dict(low=0.0,   high=1.0)),
                    scale_pos_weight        = ('uniform',    dict(low=0.01,  high=1.0))
                )

            This will be converted and used to optimize an estimator.
            """

            datasets     = model.sets
            features     = model.dp.get_estimator_features_list()
            target       = model.dp.get_target()
            search_space = model.dp.get_estimator_optimization_search_space()

            suggested_hyperparams = {
                # Convert the DataProvider.estimator_hyperparam_search_space
                # dict into Optuna´s trial.suggest_*() calls
                p: getattr(trial,'suggest_' + search_space[p][0])(
                    p,
                    **search_space[p][1]
                )
                for p in search_space
            }

            # Create placeholder similar to train dataset so we can receive
            # probas
            predicts=(
                pandas.DataFrame(
                    index=datasets['train'].index
                )
                .assign(proba=None)
            )

            for fold in self.folds:
                # Make name shortcuts
                (itrain,ival) = fold

            # for (itrain,ival) in self.skf.split(
            #             datasets['train'],
            #             datasets['train'].stratify
            #         ):

                # self.log('All params: ' + str(dict(
                #     random_state=self.random_state,
                #     **self.params,
                #     **suggested_hyperparams
                # )))

                classifier = xgboost.XGBClassifier(
                    **self.params,
                    **suggested_hyperparams
                )

                # self.log(datasets['train'].iloc[itrain].head(10).to_markdown())

                classifier.fit(
                    verbose=False,
                    X=datasets['train'].iloc[itrain][features],
                    y=datasets['train'].iloc[itrain][target],
                    eval_set=[
                        (datasets['train'].iloc[itrain][features], datasets['train'].iloc[itrain][target]),
                        (datasets['train'].iloc[ival][features],   datasets['train'].iloc[ival][target]),
                    ]
                )

                # Cirurgically set predicts in current ival rows
                predicts.proba.iloc[ival] = classifier.predict_proba(
                    datasets['train'].iloc[ival][features]
                )[:, model.dp.proba_class_index]

                # self.log(ival)
                # self.log(type(model.dp.proba_class_index))
                # self.log(classifier.predict_proba(
                #     datasets['train'].iloc[ival][features]
                # )[:, model.dp.proba_class_index])

            # auc_train is AUC between Y_true and Y_pred
            auc_train = sklearn.metrics.roc_auc_score(
                datasets['train'][target],
                classifier.predict_proba(
                        datasets['train'][features]
                    )[:, model.dp.proba_class_index]
            )

            # auc_val is AUC between Y_true and Y_val_pred
            auc_val = sklearn.metrics.roc_auc_score(
                datasets['train'][target],
                predicts.proba
            )

            return auc_train-auc_val, auc_val


        skf = sklearn.model_selection.StratifiedKFold(
            n_splits=self.bagging_size,
            shuffle=True,
            random_state=self.random_state,
        )

        self.folds = [
            itrain_ival
            for itrain_ival in skf.split(
                model.sets['train'],
                model.sets['train'].stratify
            )
        ]

        self.optimizer = optuna.create_study(
            study_name=f"{model.dp.id} • {model.get_full_train_id()}",
            directions=["minimize", "maximize"],
            storage=(
                "sqlite:///" +
                str(
                    pathlib.Path(model.get_config('TRAINED_MODELS_PATH', default='.')) /
                    f"{model.dp.id} • {model.get_full_train_id()} • optimizer.db"
                )
            ),
            load_if_exists=True,
        )

        # When Optuna 3.2.0 is widely available...
        # self.optimizer.set_metric_names([
        #     "Train AUC - Validation AUC",
        #     "Validation AUC"
        # ])

        end_by_timeout=None
        if self.optimization_timeout:
            end_by_timeout = datetime.datetime.now(
                tz=(
                    # Local timezone
                    datetime.datetime.now(datetime.timezone.utc)
                    .astimezone()
                    .tzinfo
                )
            )

            end_by_timeout += datetime.timedelta(seconds=self.optimization_timeout)

        executor=concurrent.futures.ThreadPoolExecutor()
        task=executor.submit(
            # Method name to call in the background
            self.optimizer.optimize,

            # Its parameters
            func       = lambda trial: objective(trial, model),
            n_trials   = self.optimization_trials,
            timeout    = self.optimization_timeout,
        )

        # Dump intermediary reports while waiting for the optimization to end
        while True:
            try:
                # Block until report_interval seconds. If report_interval is
                # None, block until end of optimization
                task.result(timeout=self.report_interval)
                self.logger.info('Optimization done')

                # Break away from the infinite while loop; optimization is done
                break;

            except concurrent.futures._base.TimeoutError:
                # Time to dump intermediary reports
                self.logger.info(f'Dump intermediary optimization report and object')
                self.optimizer_pareto_front = (
                    optuna.visualization.plot_pareto_front(
                        self.optimizer,
                        target_names=[
                            "Train AUC - Validation AUC",
                            "Validation AUC"
                        ]
                    )
                )

                # Compute statistics about trials and predict ETA

                trials=(
                    pandas.DataFrame(
                        [
                            dict(
                                number   = trial.number,
                                start    = trial.datetime_start,
                                complete = trial.datetime_complete,
                                state    = trial.state
                            )
                            for trial in self.optimizer.get_trials()
                            # if trial.state==optuna.trial.TrialState.COMPLETE
                        ]
                    )
                    .assign(
                        start    = lambda table: pandas.to_datetime(table.start),
                        complete = lambda table: pandas.to_datetime(table.complete),
                        duration = lambda table: table.complete-table.start
                    )
                )

                # self.logger.info(f'Trials sample:')
                # self.logger.debug(trials.to_markdown())

                if len(trials[trials.state==optuna.trial.TrialState.COMPLETE])>0:
                    now=datetime.datetime.now(
                        tz=(
                            # Local timezone
                            datetime.datetime.now(datetime.timezone.utc)
                            .astimezone()
                            .tzinfo
                        )
                    )
                    average_duration=trials[trials.state==optuna.trial.TrialState.COMPLETE].duration.median()
                    remaining_duration=(self.optimization_trials-len(trials)) * average_duration
                    eta=now + remaining_duration
                    average_duration="{}h{}m{}s".format(
                        int(average_duration.seconds/3600),
                        int((average_duration.seconds%3600)/60),
                        int((average_duration.seconds%3600)%60)
                    )
                    remaining_duration="{}h{}m{}s".format(
                        int(remaining_duration.seconds/3600),
                        int((remaining_duration.seconds%3600)/60),
                        int((remaining_duration.seconds%3600)%60)
                    )

                    message=(
                        f"{len(trials[trials.state==optuna.trial.TrialState.COMPLETE])} complete trials " +
                        f"with {average_duration} median execution time. " +
                        f"End of optimization estimated in {remaining_duration}, around {eta}."
                    )

                    if end_by_timeout:
                        message+=f" Or ends by timeout at {end_by_timeout}."

                    self.logger.info(message)

                # Let DataProviders do their things on each report_interval iteration
                if hasattr(model.dp,'post_process_after_hyperparam_optimize'):
                    model.dp.post_process_after_hyperparam_optimize(model)

        executor.shutdown()

        return self.optimizer.best_trials[0].params



    def fit_single(self, data, itrain, ival, features, target) -> xgboost.XGBClassifier:
        import sklearn

        clf = xgboost.XGBClassifier(
            **self.params,
            **self.hyperparams
        )

        # Actual training session begins

        clf.fit(
            verbose=False,
            X=data.iloc[itrain][features],
            y=data.iloc[itrain][target],
            eval_set=[
                (data.iloc[itrain][features], data.iloc[itrain][target]),
                (data.iloc[ival][features],   data.iloc[ival][target]),
            ]
        )

        return clf



    def fit(self, datasets, features, target, model=None):
        import sklearn
        # Add attribute 'max_workers=1' to inhibit parallelism

        max_workers = decouple.config('PARALLEL_ESTIMATORS_MAX_WORKERS', default=0, cast=int)
        if max_workers == '' or max_workers == 0:
            max_workers=None
            self.logger.info(f'Unlimited parallel estimators to train')
        else:
            self.logger.info(f'{max_workers} parallel estimators to train')

        skf = sklearn.model_selection.StratifiedKFold(
            n_splits=self.bagging_size,
            shuffle=True,
            random_state=self.random_state,
        )

        # Prepare infrastructure for validation data
        datasets['validation']=datasets['train']
        if model is not None and not hasattr(model, 'sets_estimations'):
            model.sets_estimations=dict()
            model.sets_estimations['validation']=None

        executor=concurrent.futures.ThreadPoolExecutor(thread_name_prefix='fit', max_workers=max_workers)
        tasks = dict()
        index = 0
        for (itrain,ival) in skf.split(
                    datasets['train'],
                    datasets['train'].stratify
                ):

            self.log(f'Trigger parallel train of XGBoost estimator #{index+1} of {self.bagging_size}...')

            task = executor.submit(
                # Method name to call asynchronously
                self.fit_single,
                # Its parameters
                datasets['train'],
                itrain,
                ival,
                features,
                target,
            )

            # Keep a track of which validation part is associated with each
            # trained XGBoost
            tasks[task]=ival

            index += 1

        self.log(f'Waiting for all member training to finish in parallel')

        # Process result as soon as it is available
        index=0
        for task in concurrent.futures.as_completed(tasks):
            xgb = task.result()
            self.bagging_members.append(xgb)

            # Collect Ŷ for validation data
            Ŷ_val = self.predict_single(
                data           = datasets['train'].iloc[tasks[task]][features],
                method         = 'predict_proba',
                bagging_member = index,
            )

            # Concatenate the validation part for this member to the whole set
            if model.sets_estimations['validation'] is None:
                model.sets_estimations['validation'] = Ŷ_val
            else:
                model.sets_estimations['validation'] = pandas.concat(
                    [
                        model.sets_estimations['validation'],
                        Ŷ_val
                    ]
                )

            index += 1
            self.log(f'Finished train of member #{index} of {self.bagging_size}…')
            self.log(f'Collecting validation data for member #{index}…')


        # Reindex all validation parts to match X.
        model.sets_estimations['validation'] = (
            model.sets_estimations['validation']
            .reindex(datasets['train'].index)
        )

        # Resulting dataframe looks like:
        # |    |   estimation_class_0 |   estimation_class_1 |   member |
        # |---:|---------------------:|---------------------:|---------:|
        # |  0 |             0.966081 |            0.0339192 |        1 |
        # |  1 |             0.95291  |            0.0470905 |        1 |
        # |  2 |             0.960247 |            0.0397534 |        2 |
        # |  3 |             0.942724 |            0.0572757 |        0 |
        # |  4 |             0.963377 |            0.036623  |        2 |

        self.log('Validation sample: \n' + model.sets_estimations['validation'].head().to_markdown())



    def predict(self, data: pandas.DataFrame) -> pandas.DataFrame:
        return self.generic_predict(data)



    def predict_proba(self, data: pandas.DataFrame, class_index: int=None) -> pandas.DataFrame:
        return self.generic_predict(
            data,
            method='predict_proba',
            class_index=class_index
        )



    def generic_predict(self, data: pandas.DataFrame, method='predict', class_index: int=None) -> pandas.DataFrame:
        """
        Predict using predict_multi() and then consolidate and aggregate estimations
        for all bagging members.
        """

        duplicate_index_error_message=[
            'DataFrame index has duplicates, which will lead',
            'to sever inconsistencies in multi-model sumarization.',
            'Here is a sample of problematic IDs; clean',
            'DataFrame appropriatelly: {}'
        ]

        duplicate_index_error_message=' '.join(duplicate_index_error_message)

        if data.shape[0]<1:
            # Empty input. Return an empty standard output.
            return pandas.DataFrame(columns=['estimation'])

        index_duplicate_count=data.index.value_counts()
        if index_duplicate_count.iloc[0] > 1:
            raise OverflowError(
                duplicate_index_error_message
                .format(list(index_duplicate_count[index_duplicate_count>1].head(10).index))
            )

        def ddebug(table,message):
            self.log(message,level=logging.DEBUG)
            return table


        if self.bagging_size>1:
            # Define aggregation strategy and compute aggregations for multiple estimations
            # in an ensamble/bagging environment.

            # Default aggregation function is mean().
            # But if this is a classifier, predict() returns a class, so we need to find the mode().
            # Aggregation strategy applies the correct aggregation function based if predict() or predict_proba()
            agg_function='mean'
            if self.is_classifier() and method == 'predict':
                agg_function=pandas.Series.mode

            agg_strategy=dict(estimation=agg_function)
            if method == 'predict_proba' and class_index is None:
                agg_strategy={
                    f'estimation_class_{col}':agg_function
                    for col in self.bagging_members[0].classes_
                }

            self.log(f"Bagging aggregation strategy: {agg_strategy}",level=logging.DEBUG)

            return (
                self.predict_multi(data,method,class_index)

                # Group by index
                .groupby(level=0)

                # Member ID was consolidated and ghosted at this point.

                # Aggregated results by mean or mode (only if classifier and predict method).
                # One of the reasons Index can’t have duplicates is to avoid
                # bias in here.
                .agg(agg_strategy)

                # Order DataFrame the same as input
                .reindex(data.index)
            )
        else:
            return self.predict_single(data,method,class_index).drop(columns=['member'])



    ####################################################################################
    ###
    ### Internal methods
    ###
    ####################################################################################



    def predict_single(self, data: pandas.DataFrame, method='predict', class_index: int=None, bagging_member: int=0) -> pandas.DataFrame:
        """
        Return raw estimations of one estimator member tagged by member ID.
        If method is 'predict_proba' all class probabilities are returned in multiple
        columns called estimation_class_{I}. Unless class_index is passed, then only
        that class is returned.
        
        Resulting dataframe has columns:

        - index
        - estimator member ID (0,1,2,3 etc as passed in bagging_member)
        - estimation (if predict) or estimation_class_{I} (if predict_proba)

        This method was designed to be parallelized with concurrent.futures.
        """
        
        self.log(f'Member #{bagging_member} is predicting for {data.shape[0]} datapoints...',level=logging.DEBUG)
        
        def ddebug(table,message):
            self.log(message,level=logging.DEBUG)
            return table
        
        renamer=lambda col: 'estimation'
        unwanted_classes=list()
        
        if method == 'predict_proba':
            if class_index is not None:
                # Want single value, not probabilities of all classes
                unwanted_classes=list(set(self.bagging_members[bagging_member].classes_)-set([class_index]))
            else:
                # Change strategy for column renamer
                renamer=lambda col: f'estimation_class_{col}'
        
        return (
            pandas.DataFrame(
                index=data.index,
                
                # Compute estimation via predict or predict_proba, using 1 member
                data=getattr(self.bagging_members[bagging_member], method)(data)
            )
            
            # Keep only the class_index class, if provided
            .drop(columns=unwanted_classes)
            
            # Rename columns from 0, 1, etc to estimation (if predict) or estimation_class_{I}
            .rename(columns=renamer)
            
            # Tag it with member ID
            .assign(member=bagging_member)
            
            # Reduce RAM usage
            .assign(
                member=lambda table: table.member.astype('category')
            )
            
            .pipe(lambda table: ddebug(table,f'Member #{bagging_member} finished predicting for {data.shape[0]} datapoints.'))
        )



    def predict_multi(self, data: pandas.DataFrame, method='predict', class_index: int=None) -> pandas.DataFrame:
        """
        Return raw estimations of all estimator members tagged by member ID in
        a MultiIndex DataFrame.

        Resulting dataframe has columns:

        Index:
        - (original index, preferably unit_id)
        - estimator member ID (0,1,2,3 etc)

        Columns:
        - estimation (if predict) or estimation_class_{I} (if predict_proba)
        """

        executor=concurrent.futures.ThreadPoolExecutor(thread_name_prefix='pred_multi_dist')
        estimation=None
        tasks=[]

        # Trigger all estimators in parallel
        for i in range(self.bagging_size):
            tasks.append(
                # Execute in the background ...
                executor.submit(
                    # ... this method ...
                    self.predict_single,

                    # ... first parameter ...
                    data,

                    # ... second parameter ...
                    method,

                    # ... third parameter ...
                    class_index,

                    # ... fourth parameter ...
                    i,
                )
            )

        # Process their resulting DataFrame as soon as it is available
        for task in concurrent.futures.as_completed(tasks):
            if estimation is None:
                estimation=task.result()
            else:
                estimation=pandas.concat([estimation,task.result()])

        return estimation.set_index([estimation.index,'member']).sort_index()



    def __getstate__(self):
        import json

        bagging_members_safe = list()

        # Array of trained XGBoosts in a format that can be serialized
        for x in self.bagging_members:
            with tempfile.NamedTemporaryFile(suffix='.json', mode='w+t') as tmp:
                x.save_model(tmp.name)
                bagging_members_safe.append(json.load(tmp))

        return dict(
            **xingu.Estimator.__getstate__(self),

            # Number of trained models. Should be same as len(bagging_members)
            bagging_size     = self.bagging_size,

            bagging_members_safe = bagging_members_safe,
            # bagging_members  = self.bagging_members,

            # Random number used by the class, as 42
            random_state     = self.random_state,

            # xgboost_missing_classes_ = self.bagging_members[0].classes_,
        )



    def __setstate__(self, state):
        import json

        self.__dict__.update(state)

        self.bagging_members=list()
        for serialized in self.bagging_members_safe:
            with tempfile.NamedTemporaryFile(suffix='.json', mode='w+', delete=False) as tmp:
                json.dump(serialized, tmp)

            x=xgboost.XGBClassifier()
            x.load_model(tmp.name)
            self.bagging_members.append(x)

            pathlib.Path(tmp.name).unlink()

        del self.bagging_members_safe



    def is_classifier(self):
        from sklearn.base import is_classifier
        return is_classifier(self.bagging_members[0])


__all__ = ['Model', 'Coach', 'Estimator', 'DataProviderFactory', 'DataProvider', 'DPSimple', 'ConfigManager']

from .dataprovider          import DataProvider
from .dataprovider          import DPSimple
from .config_manager        import ConfigManager
from .dataproviderfactory   import DataProviderFactory
from .estimator             import Estimator
from .coach                 import Coach
from .model                 import Model



import collections
import concurrent.futures
import copy
import datetime
import urllib
import importlib
import logging
import socket
import pathlib
import queue
import random
import os

import pandas
import yaml

from . import DataProvider
from . import DataProviderFactory
from . import Estimator
from . import ConfigManager



class Coach:

    ######################################################################################
    ##
    ## Initialization and configuration
    ##
    ######################################################################################


    defaults=dict(
        HYPEROPT_STRATEGY       = 'last',
        TRAINED_MODELS_PATH     = None,
        DVC_TRAINED_MODELS_PATH = None,
        QUERY_CACHE_PATH        = None,
        DVC_QUERY_CACHE_PATH    = None,
        XINGU_DB_TABLE_PREFIX  = '',
        DATA_DB_URL             = None,
        PROJECT_HOME            = '.',
    )



    def __init__(self, dp_factory: DataProviderFactory = DataProviderFactory()):
        self.config=ConfigManager()

        # Setup logging
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)

        # Turn attribute `databases` into a private copy
        self.databases=dict()

        try:
            import pygit2
            self.git_repo=pygit2.Repository(self.get_config('PROJECT_HOME'))
            self.logger.debug("Using git repo on {}".format(self.git_repo))
        except:
            self.git_repo=None

        self.dp_factory=dp_factory

        # Load our map of DataProvider IDs and their current train_ids
        self.load_inventory()

        # Database initialization block
        self.tables={}
        self.xingu_db=None

        # Flow control on queues and parallelism
        self.post_processing=False



    def get_config(self, config_item: str, default=ConfigManager.undefined, cast=ConfigManager.undefined):
        if default==ConfigManager.undefined:
            if config_item in self.defaults:
                return self.config.get(config_item, default=self.defaults[config_item], cast=cast)
            else:
                return self.config.get(config_item, cast=cast)
        else:
            return self.config.get(config_item, default=default, cast=cast)



    ######################################################################################
    ##
    ## Team train and team batch predict
    ##
    ## Methods are called following this topology:
    ##
    ## team_train():
    ##   - team_train_parallel() (background, parallelism controled by PARALLEL_TRAIN_MAX_WORKERS):
    ##     - team_load() (for pre-req models not trained in this session)
    ##     - Per DataProvider requested to be trained:
    ##       - team_train_member() (background):
    ##         - Model.fit()
    ##   - post_train_parallel() (background, only if POST_PROCESS=true):
    ##     - Per trained estimator (parallelism controled by PARALLEL_POST_PROCESS_MAX_WORKERS):
    ##       - Model.save() (background)
    ##       - Model.trainsets_save() (background)
    ##       - single_batch_predict() (background, only if bacth_predict=True):
    ##         - Model.compute_and_save_metrics()
    ##         - Model.save_batch_predict_estimations()
    ##   - update_inventory()
    ##
    ##
    ##
    ##
    ## If in batch-predict-only mode (use pre-trained estimators):
    ##
    ## team_batch_predict():
    ##   - team_load() (for all requested DPs)
    ##   - Per loaded model:
    ##     - single_batch_predict() (background):
    ##       - Model.compute_and_save_metrics()
    ##       - Model.save_batch_predict_estimations()
    ##
    ##
    ######################################################################################



    def trained_ids(self):
        """
        Return a set of just the IDs of what we have in our inventory of trained Models
        """
        return self.trained.keys()



    def team_train_member(self,dp):
        """
        Train a single Robson for DataProvider dp and add it to inventory
        """
        # Imported here to avoid circular dependency problems
        from xingu import Model

        # Actual model training
        model=Model(
            dp                   = dp,
            coach                = self,
            trained              = False,
            hyperopt_strategy    = self.get_config('HYPEROPT_STRATEGY')
        )

        self.logger.info(f'Start training {model.dp.id}')
        model.fit() # a lengthy process
        self.logger.info(f'Finished training {model.dp.id}')

        # Add to inventory of trained models
        self.declare_trained(model)



    def team_train_parallel(self):
        import randomname

        self.train_session_id=self.get_config('TRAIN_SESSION_ID', default=None)
        if self.train_session_id is None:
            self.train_session_id=randomname.get_name()

        self.trained = {}
        self.trained_in_session = []

        self.logger.info('Training Models for the following DataProviders: ' + str(self.dp_factory.providers_list))


        ##################################################################################
        ##
        ## Compute which pre-trained models need to be loaded in order to train what
        ## we were requested to train. Load pre-trained pre-reqs with team_load()
        ##
        ##################################################################################

        to_train = list(self.dp_factory.produce())

        to_load=set()
        for dp in to_train:
            to_load = to_load | self.dp_factory.get_pre_req(dp)

        # Compute list of pre-req models that need to be loaded except the ones
        # that are going to be trained
        to_load = to_load - set([dp.id for dp in to_train])

        if len(to_load):
            # Check if we have special order requirements for pre-reqs
            pre_req_train_or_session_ids=self.get_config('PRE_REQ_TRAIN_OR_SESSION_IDS', default=None)
            if pre_req_train_or_session_ids is not None:
                # Convert to a list
                pre_req_train_or_session_ids=[x.strip() for x in pre_req_train_or_session_ids.split(',')]

            self.logger.info('Loading pre-trained models as pre-reqs: ' + str(to_load))
            self.team_load(
                to_load,
                pre_req_train_or_session_ids=pre_req_train_or_session_ids,
                post_process=False
            )

        random.shuffle(to_train)
        to_train=collections.deque(to_train)

        # Define how many parallel workers to execute
        max_workers=self.get_config('PARALLEL_TRAIN_MAX_WORKERS', default=0, cast=int)
        if max_workers == '' or max_workers == 0:
            max_workers=None
            self.logger.info(f'Training all possible Models in parallel')
        else:
            self.logger.info(f'Training {max_workers} Models in parallel')


        with concurrent.futures.ThreadPoolExecutor(
                            thread_name_prefix='team_train_parallel',
                            max_workers=max_workers
                ) as executor:
            processing=[]
            waiting=[]

            while len(to_train):
                # First try to submit all
                while len(to_train):
                    dp=to_train.popleft()
                    if hasattr(dp,'pre_req'):
                        if len(dp.pre_req) == len(dp.pre_req.intersection(self.trained_ids())):
                            self.logger.info('Trigger train of «{}»'.format(dp.id))
                            processing.append(executor.submit(self.team_train_member,dp))
                        else:
                            # Not ready for you yet. Go back to the queue.
                            self.logger.info('Deferring train of «{}» until all its pre-reqs are ready'.format(dp.id))
                            waiting.append(dp)
                    else:
                        # Model has no dependencies, so just train it
                        self.logger.info('Trigger train of «{}»'.format(dp.id))
                        processing.append(executor.submit(self.team_train_member,dp))

                # Requeue everything that we couldn't process due to pre-reqs
                to_train.extend(waiting)
                waiting=[]

                # Now block until something finishes
                tasks=concurrent.futures.wait(processing, return_when=concurrent.futures.FIRST_COMPLETED)
                for task in tasks.done:
                    # Let finished tasks express themselves. This is where in-task
                    # exceptions are raised.
                    e=task.exception()
                    processing.remove(task)
                    if e is None:
                        task.result()
                    else:
                        self.logger.warning('Exception occurred in team_train_parallel() tasks.')
                        for t in tasks.not_done:
                            # Cancel everything that we can cancel
                            t.cancel()
                        self.logger.exception(e)
                        raise e

        # Put an empty object to flag the end of queue
        if self.post_train_queue is not None:
            self.post_train_queue.put(None, block=False)



    def declare_trained(self, model, post_process=True):
        # Add to inventory of trained models
        self.trained[model.dp.id]=model

        # Add to inventory of models trained in this train session
        if hasattr(self,'trained_in_session'):
            self.trained_in_session.append(model.dp.id)

        if post_process and self.post_processing:
            # Put on a queue for post-process tasks, because
            # post_train_parallel() is waiting.
            self.post_train_queue.put(model, block=False)



    def team_train(self):
        """
        Train and post-process Models for all DPs requested.

        After an estimator is trained, it enters the post-process phase, which
        executes the following:
        1. Save PKL and commit to DVC
        2. Save Train, Validation and Test datasets to DB
        3. Get data and execute Batch Predict
        4. Save estimations in valuations DB table
        5. Compute various metrics and save them to DB.

        If POST_PROCESS env variable is set and is false, none of this will
        happen and this is usualy the behavior when we are only optimizing
        hyper-parameters.

        If POST_PROCESS env variable is not set or is true, post-processing
        activities are executed.

        If BATCH_PREDICT==False, steps from 3 and beyond are not executed.

        The BATCH_PREDICT flag is ignored if POST_PROCESS is false.
        """

        self.post_processing=False
        self.post_train_queue=None

        do_post_process=self.get_config('POST_PROCESS', default=True, cast=bool)

        with concurrent.futures.ThreadPoolExecutor(thread_name_prefix='team_train') as executor:
            # Only 2 tasks here:
            # - Trigger all trains in parallel, taking care of pre-reqs
            # - As soon as trained Models are ready, do their post processing
            tasks=[]

            # The parallel post-train task, including batch predict
            if do_post_process:
                # Initialize a queue for post-train activities
                self.post_train_queue=queue.Queue()

                # Let everybody know we are post-processing trained Models
                self.post_processing=True

                # Start post-processing
                tasks.append(executor.submit(self.post_train_parallel))

            # The parallel train task
            tasks.append(executor.submit(self.team_train_parallel))

            # Wait for both tasks (train and post-process) to finish
            for task in concurrent.futures.as_completed(tasks):
                # Let finished tasks express themselves. This is where in-task
                # exceptions are raised.
                e=task.exception()
                if e is None:
                    task.result()
                else:
                    self.logger.warning('Exception ocurred in team_train() tasks. Forcing a shutdown in post-process task.')
                    self.post_processing=False
                    if self.post_train_queue is not None:
                        self.post_train_queue.put(None, block=False)
                        self.logger.debug('Post process queue size: ' + str(self.post_train_queue.qsize()))
                    self.logger.exception(e)
                    raise e

            self.post_processing=False

        # Block until everything is done in this post-train queue
        if self.post_train_queue is not None:
            self.logger.info('Wait for post-process to finish while {} tasks in queue.'.format(self.post_train_queue.qsize()))
            self.post_train_queue.join()

        for model in self.trained_in_session:
            self.trained[model].cleanup()
            # If DVC is active, there will be things to commit
            self.trained[model].dvc_commit()

        self.update_inventory()



    def post_train_parallel(self):
        """
        Listens to trained Models and fire post-process tasks such as saving model,
        saving data, batch predict, metrics computation etc.
        """

        # Define how many parallel workers to execute
        max_workers=self.get_config('PARALLEL_POST_PROCESS_MAX_WORKERS', default=0, cast=int)
        if max_workers == '' or max_workers == 0:
            max_workers=None
            self.logger.info(f'Post-processing all possible Models in parallel')
        else:
            self.logger.info(f'Post-processing no more than {max_workers} Models in parallel')


        with concurrent.futures.ThreadPoolExecutor(
                            thread_name_prefix='post_train_parallel',
                            max_workers=max_workers
                ) as executor:
            tasks=[]
            while True:
                # Block until a Model object is trained and ready to be used for post processing
                model=self.post_train_queue.get()
                self.post_train_queue.task_done()

                self.logger.debug('Post process queue size after queue.get(): ' + str(self.post_train_queue.qsize()))

                if model is None:
                    # A sign that all was done. Exit the endless loop.
                    self.post_processing=False
                    self.logger.info('All post-process tasks done. Shutting down queue.')
                    break

                self.logger.info(f'A fresh Model({model.dp.id}) is ready for post-processing')


                # Now do all the amazing things we can do with a trained
                # Model. In parallel.

                ## Save pickle to storage
                tasks.append(
                    executor.submit(
                        model.save,
                        **dict(
                            path      = self.get_config(    'TRAINED_MODELS_PATH'),
                            dvc_path  = self.get_config('DVC_TRAINED_MODELS_PATH'),
                        )
                    )
                )

                if self.get_config('SAVE_SETS'):
                    # Save trainsets to DB
                    tasks.append(executor.submit(model.trainsets_save))

                # Predict and compute metrics over train data
                tasks.append(executor.submit(model.trainsets_predict))

                # Batch predict and metrics computation
                if self.get_config('BATCH_PREDICT', default=False, cast=bool):
                    tasks.append(executor.submit(self.single_batch_predict,model))
                else:
                    self.logger.info('Skipping batch predict activities.')


            for task in concurrent.futures.as_completed(tasks):
                # Let finished tasks express themselves. This is where
                # in-task exceptions are raised.

                self.logger.debug('A post-process task is done.')

                e=task.exception()
                if e is None:
                    # No exception
                    task.result()
                else:
                    self.logger.warning('Exception ocurred in team_train() tasks. Forcing a shutdown in post-process task.')
                    self.post_processing=False
                    self.post_train_queue.put(None, block=False)
                    self.logger.exception(e)
                    raise e

        self.logger.debug('Alls post-process tasks finished.')



    def single_batch_predict(self, model):
        model.batch_predict()

        # For each trained Robson, do 2 tasks in parallel
        with concurrent.futures.ThreadPoolExecutor(thread_name_prefix='single_batch_predict') as executor:
            tasks=[]

            # Task 1: compute and save metrics
            tasks.append(executor.submit(model.compute_and_save_metrics,'batch'))

            # Task 2: save estimations
            if self.get_config('BATCH_PREDICT_SAVE_ESTIMATIONS', default=False, cast=bool):
                tasks.append(executor.submit(model.save_batch_predict_estimations))

            for task in concurrent.futures.as_completed(tasks):
                # Does nothing if thread succeeded. Raises the task's exception otherwise.
                task.result()



    def team_batch_predict(self):
        if not hasattr(self, 'trained'):
            self.team_load()

        # Define how many parallel workers to execute
        max_workers=self.get_config('PARALLEL_POST_PROCESS_MAX_WORKERS', default=0, cast=int)
        if max_workers == '' or max_workers == 0:
            max_workers=None
            self.logger.info(f'Using all possible Models in parallel')
        else:
            self.logger.info(f'Using no more than {max_workers} Models in parallel')

        # All models in place, so now batch predict in parallel
        with concurrent.futures.ThreadPoolExecutor(
                            thread_name_prefix='team_batch_predict',
                            max_workers=max_workers
                ) as executor:
            tasks=[]

            for model in self.trained:
                tasks.append(executor.submit(self.single_batch_predict, self.trained[model]))

            for task in concurrent.futures.as_completed(tasks):
                # Does nothing if thread succeeded. Raises the task's exception otherwise.
                task.result()



    def update_inventory(self):
        if hasattr(self,'trained') and hasattr(self,'trained_in_session'):
            # Make any needed initialization
            if 'history' not in self.inventory:
                self.inventory['history']=[]

            if 'models' not in self.inventory:
                self.inventory['models']=dict()

            # Update file with what we have
            self.inventory['history'].append(
                dict(
                    time              = datetime.datetime.now(datetime.timezone.utc),
                    user_name         = self.get_config('USERNAME',os.environ.get('USER', os.environ.get('USERNAME'))),
                    host_name         = self.get_config('HOSTNAME',socket.gethostname()),
                    git_branch        = self.git_repo.head.name if self.git_repo else None,
                    git_commit        = self.git_repo.head.target.hex if self.git_repo else None,
                    github_actor      = self.get_config('GITHUB_ACTOR', None),
                    github_workflow   = self.get_config('GITHUB_WORKFLOW', None),
                    github_run_id     = self.get_config('GITHUB_RUN_ID', None),
                    github_run_number = self.get_config('GITHUB_RUN_NUMBER', None),
                    trained_in_session= self.trained_in_session,
                    train_session_id  = self.train_session_id
                )
            )

            # Preserve just last 10 entries in history
            self.inventory['history']=self.inventory['history'][-10:]

            for e in self.trained_in_session:
                if e in self.inventory['models']:
                    del self.inventory['models'][e]
                self.inventory['models'][e]=self.trained[e].signature()

            target=pathlib.Path(self.get_config('PROJECT_HOME')).resolve() / 'inventory.yaml'
            with open(target, 'w') as f:
                yaml.dump(
                    data=self.inventory,
                    stream=f,

                    # YAML styling options
                    explicit_start=True,
                    canonical=False,
                    indent=6,
                    default_flow_style=False
                )

            self.git_commit_inventory()



    def load_inventory(self) -> dict:
        target=pathlib.Path(self.get_config('PROJECT_HOME')).resolve() / 'inventory.yaml'

        try:
            self.logger.debug("Trying to open {}".format(target))
            with open(target) as f:
                self.inventory = yaml.safe_load(f)
        except FileNotFoundError as e:
            self.inventory={}

        if self.inventory is None:
            self.inventory={}

        return self.inventory



    def git_commit_inventory(self):
        if self.get_config('COMMIT_INVENTORY', default=False, cast=bool):
            git_shell = """
                cd {path};
                git add 'inventory.yaml';
                git commit -m 'chore(inventory.yaml)';
                git push;
            """

            git_shell=git_shell.format(path=pathlib.Path(self.get_config('PROJECT_HOME')).resolve())
            self.logger.debug(f'Commit inventory.yaml to Git with command:\n{git_shell}')
            return_code=os.system(git_shell)

            if return_code != 0:
                raise OSError(return_code, f'Following external command failed: {git_shell}')


    ######################################################################################
    ##
    ## Services for team coordination
    ##
    ######################################################################################


    def find_model_for_dp(self, dp: str):
        if hasattr(self,'trained'):
            return self.trained[dp]



    def team_load(
                    self,
                    explicit_list                              = None,
                    post_process:                         bool = True,
                    estimator_only:                       bool = False,
                    pre_trained_path:                     str  = None,
                    pre_trained_train_session_id:         str  = '*',
                    pre_trained_as_of:                    str  = '*',
                    pre_req_train_or_session_ids:         list = None,
                    xingu_module:                         str  = 'xingu',
                    model_class:                          str  = 'Model'
        ):
        """
        Load pre-trained Xingu models into Coach’s inventory.

        PKLs will be searched in S3 or filesystem path defined in
        pre_trained_path.

        explicit_list -- Load estimators only for this DataProviders. Load all
        if empty or None.

        post_process -- If True, submit loaded models for post-processing
        (batch predict, metrics computation etc)
        """
        xingu_module = importlib.import_module(xingu_module)
        Model = getattr(xingu_module, model_class)

        model_params_template=dict(
            # Our optimization technique
            delayed_prereq_binding=               True,

            # Obvious parameter
            trained=                              True,

            # Other stuff
            estimator_only=                       estimator_only,
            pre_trained_path=                     pre_trained_path,
            pre_trained_train_session_id=         pre_trained_train_session_id,
            pre_trained_train_id=                 '*',
            pre_trained_as_of=                    pre_trained_as_of,
            train_or_session_id_match_list=       pre_req_train_or_session_ids
        )

        # Start with a fresh empty team
        self.trained = dict()

        if explicit_list is None:
            explicit_list=list(self.dp_factory.produce())

        # Find all pre-reqs of desired models and add them to the list
        to_load = set()
        for dp in explicit_list:
            if isinstance(dp, str):
                to_load=to_load | {dp}    | self.dp_factory.get_pre_req(dp)
            else:
                to_load=to_load | {dp.id} | self.dp_factory.get_pre_req(dp)

        self.logger.debug(f'Going to load pre-trained: {to_load}')

        # Parallel load of all desired pre-trained Robson objects
        with concurrent.futures.ThreadPoolExecutor(thread_name_prefix='team_load') as executor:
            tasks = []

            for dp in to_load:
                model_params = copy.deepcopy(model_params_template)
                model_params['coach'] = self
                model_params['dp'] = self.dp_factory.dps[dp]() if isinstance(dp, str) else dp

                if (
                            'models' in self.inventory and
                            dp in self.inventory['models']
                ):
                    model_params['pre_trained_train_id'] = self.inventory['models'][dp]['train_id']
                    model_params['train_or_session_id_match_list'] = None

                self.logger.info(f'Trigger background loading of ‘{dp}’')
                tasks.append(executor.submit(Model, **model_params))

            for task in concurrent.futures.as_completed(tasks):
                # Does nothing if thread succeeded. Raises the task's exception otherwise.
                mod = task.result()
                self.declare_trained(mod, post_process=post_process)
                self.logger.info(f'Loaded {mod}')

        # Now bind pre-req models with the ones already in the team.
        for model in self.trained:
            self.trained[model].load_pre_req_model()



    ######################################################################################
    ##
    ## Metrics report
    ##
    ######################################################################################


    def report(self, train_ids: list=None, dataprovider_id: str=None, on: str='model',
                    start: str=None, reference_train_id: str=None) -> pandas.DataFrame:

        import sqlalchemy

        # Method idealized but still unimplemented.

        metric_table            = self.tables['metrics_' + on]
        training_table          = self.tables['training']
        training_status_table   = self.tables['training_status']

        query_meta=(
            training_table.select()
                .join(training_status_table, sqlalchemy.and_(
                        training_status_table.c.dataprovider_id == training_table.c.dataprovider_id,
                        training_status_table.c.train_id        == training_table.c.train_id
                    )
                )
                .filter(training_status_table.c.status == 'train_fit_end')

                .with_only_columns(
                    training_status_table.c.time,
                    training_table.c.dataprovider_id,
                    training_table.c.train_session_id,
                    training_table.c.train_id,
                    training_table.c.user_name,
                    training_table.c.host_name,
                    training_table.c.git_branch,
                    training_table.c.git_commit,
                    training_status_table.c.hyperparam
                )
        )

        if on == 'model':
            column_set='time_utc dataprovider_id train_session_id train_id set name'.split()
            time_table=training_status_table

            query_metrics=(
                metric_table.select()
                    .join(training_status_table, sqlalchemy.and_(
                            training_status_table.c.dataprovider_id == metric_table.c.dataprovider_id,
                            training_status_table.c.train_id        == metric_table.c.train_id
                        )
                    )
                    .filter(training_status_table.c.status == 'model_metrics_end')

                    .with_only_columns(
#                         training_status_table.c.time,
#                         metric_table.c.dataprovider_id,
#                         metric_table.c.train_session_id,
                        metric_table.c.train_id,
                        metric_table.c.set,
                        metric_table.c.name,
                        metric_table.c.value_number,
                        metric_table.c.value_text,
                    )
            )

        elif on == 'valuation':
            column_set='unit_id dataprovider_id time_utc train_id name'.split()
            time_table=metric_table

            query_metrics=(
                metric_table.select()
                    .with_only_columns(
                        metric_table.c.unit_id,
                        metric_table.c.dataprovider_id,
                        metric_table.c.time,
                        metric_table.c.train_session_id,
                        metric_table.c.train_id,
                        metric_table.c.name,
                        metric_table.c.value_number,
                        metric_table.c.value_text,
                    )
            )


        if dataprovider_id is not None:
            query_metrics = query_metrics.filter(metric_table.c.dataprovider_id == dataprovider_id)
            query_meta    = query_meta.filter(training_table.c.dataprovider_id == dataprovider_id)

        all_train_ids=[]
        if isinstance(train_ids, list):
            all_train_ids=train_ids

        if reference_train_id is not None:
            all_train_ids.append(reference_train_id)

        if len(all_train_ids)>0:
            query_metrics = query_metrics.filter(metric_table.c.train_id.in_(all_train_ids))
            query_meta    = query_meta.filter(training_table.c.train_id.in_(all_train_ids))

        if start is not None:
            query_metrics = query_metrics.filter(time_table.c.time >= pd.Timestamp(start).timestamp())
            query_meta    = query_meta.filter(training_status_table.c.time >= pd.Timestamp(start).timestamp())


        # Now execute query
        report={}
        self.logger.debug(
            'Estimator metadata retrieval query: ' +
            query_meta.compile(compile_kwargs={'literal_binds': True}).string
        )
        self.logger.debug(
            'Metric retrieval query: ' +
            query_metrics.compile(compile_kwargs={'literal_binds': True}).string
        )
        report['meta']    = pd.read_sql(query_meta,    con=query_meta.bind)
        report['metrics'] = pd.read_sql(query_metrics, con=query_metrics.bind)

        # Handle time
        report['meta']['time_utc']=pd.to_datetime(report['meta']['time'], unit='s', utc=True)
        report['meta']=report['meta'].set_index('train_id').drop(columns=['time']).T

        report['metrics']['value'] = (
            report['metrics']['value_number']
            .combine_first(report['metrics']['value_text'])
        )

        report['metrics']=(
            report['metrics']
            .drop(columns=['value_number','value_text'])
            .set_index(['name','set','train_id'])
            .unstack(level=['train_id','set'])
        )


        return report

        df=df.set_index(column_set).sort_index().unstack()

        for c in df.columns:
            if df[c].value_counts().shape[0]==0:
                # Detected a column containing nothing but NaNs, a residual from unstack(); remove it
                df.drop(columns=[c], inplace=True)
        df.columns=df.columns.droplevel(0)

        df.sort_index(inplace=True)

        return df



    ######################################################################################
    ##
    ## Database management and creation
    ##
    ######################################################################################


    def get_db_connection(self, nickname='xingu'):
        if nickname in self.databases:
            return self.databases[nickname]['conn']

        import sqlalchemy

        engine_config_sets=dict(
            # Documentation for all these SQLAlchemy pool control parameters:
            # https://docs.sqlalchemy.org/en/14/core/engines.html#engine-creation-api

            DEFAULT=dict(
                # QueuePool config for a real database
                poolclass         = sqlalchemy.pool.QueuePool,

                # 5 is the default. Reinforce default, which is good
                pool_size         = 5,

                # Default here was 10, which might be low sometimes, so
                # increase to some big number in order to never let the
                # QueuePool be a bottleneck.
                max_overflow      = 50,

                # Debug connection and all queries
                # echo              = True
            ),
            sqlite=dict(
                # SQLite doesn’t support concurrent writes, so we‘ll amend
                # the DEFAULT configuration to make the pool work with only
                # 1 simultaneous connection. Since Robson is agressively
                # parallel and requires a DB service that can be used in
                # parallel (regular DBs), the simplicity and portability
                # offered by SQLite for a light developer laptop has its
                # tradeoffs and we’ll have to tweak it to make it usable in
                # a parallel environment even if SQLite is not parallel.

                # A pool_size of 1 allows only 1 simultaneous connection.
                pool_size         = 1,
                max_overflow      = 0,

                # Since we have only 1 stream of work (pool_size=1),
                # we need to put a hold on other DB requests that arrive
                # from other parallel tasks. We do this putting a high value
                # on pool_timeout, which controls the number of seconds to
                # wait before giving up on getting a connection from the
                # pool.
                pool_timeout      = 3600.0,

                # Debug connection and all queries
                # echo              = True
            ),
        )

        # Convert the text on environment ("nick1|url1|nick2|url2") into
        # a clean list
        databases=self.get_config('DATABASES', default=None)

        if databases is None:
            databases=list()
        else:
            databases=[i for i in [i.strip() for i in databases.split('|')] if i != '']

        # Process Xingu DB likewise with nickname 'xingu'
        if nickname == 'xingu':
            databases += ['xingu',self.get_config('XINGU_DB_URL')]

        if (len(databases) % 2) != 0:
            raise NotImplementedError('Malformed DATABASES environment. Format is "nick1|url1|nick2|url2".')

        for i in range(0,len(databases),2):
            self.databases[databases[i]] = dict(
                url=databases[i+1]
            )

            current=self.databases[databases[i]]

            # Start with a default config
            engine_config=engine_config_sets['DEFAULT'].copy()

            # Add engine-specific configs
            for dbtype in engine_config_sets.keys():
                # Extract from engine_config_sets configuration specific
                # for each DB type
                if dbtype in current['url']:
                    engine_config.update(engine_config_sets[dbtype])

            # Databricks needs special URL handling
            # URLs are like "databricks+connector://host.com/default?http_path=/sql/..."
            if 'databricks' in current['url']:
                tokenized = urllib.parse.urlparse(current['url'])

                # Extract connection args as "?http_path=..." into a dict
                # Returns {'http_path': ['value']}
                conn_args = urllib.parse.parse_qs(tokenized.query)

                # Unwrap the value from the list
                # Return  {'http_path': 'value'}
                engine_config.update(
                    dict(
                        connect_args={k:conn_args[k][0] for k in conn_args}
                    )
                )

                # Reconstruct the URL without the connection args
                tokenized=tokenized._replace(query=None)
                current['url']=urllib.parse.urlunparse(tokenized)

            current['conn'] = sqlalchemy.create_engine(
                url = current['url'],
                **engine_config
            )

            # DB Engine was just created; check if all tables are there
            if databases[i] == 'xingu':
                self.init_db(self.databases[databases[i]])

            self.logger.debug(f"Data source «{databases[i]}» is {current['conn']}. Created with config: {engine_config}.")

        return self.databases[nickname]['conn']



    def init_db(self, con):
        if self.get_config('XINGU_DB_URL'):
            # This is just to raise an exception if not set.
            # Can't do Coach business without a DB.
            pass

        import sqlalchemy

        self.xingu_db_metadata = sqlalchemy.MetaData(bind=con['conn'])

        self.xingu_db_table_prefix=self.get_config('XINGU_DB_TABLE_PREFIX','')

        table_name='training'
        self.tables[table_name] = sqlalchemy.Table(
            self.xingu_db_table_prefix + table_name,
            self.xingu_db_metadata,
            sqlalchemy.Column(
                'train_session_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'dataprovider_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),

            sqlalchemy.UniqueConstraint(
                'train_session_id',
                'train_id',
                'dataprovider_id',
                name='train'
            )
        )

        table_name='training_attributes'
        self.tables[table_name] = sqlalchemy.Table(
            self.xingu_db_table_prefix + table_name,
            self.xingu_db_metadata,
            sqlalchemy.Column(
                'train_session_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'dataprovider_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'attribute', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'value', sqlalchemy.String
            ),

            sqlalchemy.UniqueConstraint(
                'train_session_id',
                'train_id',
                'dataprovider_id',
                'attribute',
                name='train'
            )
        )

        table_name='training_steps'
        self.tables[table_name] = sqlalchemy.Table(
            self.xingu_db_table_prefix + table_name,
            self.xingu_db_metadata,
            sqlalchemy.Column(
                'time', sqlalchemy.Integer,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'dataprovider_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_session_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'status', sqlalchemy.String,
                primary_key=True
            ),

            sqlalchemy.ForeignKeyConstraint(
                [
                    'dataprovider_id',
                    'train_session_id',
                    'train_id'
                ],
                [
                    self.xingu_db_table_prefix + 'training.dataprovider_id',
                    self.xingu_db_table_prefix + 'training.train_session_id',
                    self.xingu_db_table_prefix + 'training.train_id',
                ]
            ),
        )

        if 'awsathena' not in con['url']:
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_time',             self.tables[table_name].time),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_dataprovider_id',  self.tables[table_name].dataprovider_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_train_session_id', self.tables[table_name].train_session_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_train_id',         self.tables[table_name].train_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_status',           self.tables[table_name].status),


        table_name='sets'
        self.tables[table_name] = sqlalchemy.Table(
            self.xingu_db_table_prefix + table_name,
            self.xingu_db_metadata,
            sqlalchemy.Column(
                'dataprovider_id', sqlalchemy.String,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_session_id', sqlalchemy.String,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_id', sqlalchemy.String,
                nullable=False
            ),
            sqlalchemy.Column(
                'set', sqlalchemy.String,
                nullable=False
            ),
            sqlalchemy.Column(
                'index', sqlalchemy.String,
                nullable=False
            ),
            sqlalchemy.Column('target', sqlalchemy.Float),

            sqlalchemy.ForeignKeyConstraint(
                [
                    'dataprovider_id',
                    'train_session_id',
                    'train_id'
                ],
                [
                    self.xingu_db_table_prefix + 'training.dataprovider_id',
                    self.xingu_db_table_prefix + 'training.train_session_id',
                    self.xingu_db_table_prefix + 'training.train_id',
                ]
            ),
        )

        if 'awsathena' not in con['url']:
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_dataprovider_id',  self.tables[table_name].dataprovider_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_train_session_id', self.tables[table_name].train_session_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_train_id',         self.tables[table_name].train_id),


        table_name='metrics_model'
        self.tables[table_name] = sqlalchemy.Table(
            self.xingu_db_table_prefix + table_name,
            self.xingu_db_metadata,
            sqlalchemy.Column(
                'time', sqlalchemy.Integer,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'dataprovider_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_session_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'set', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'name', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column('value_number', sqlalchemy.Float),
            sqlalchemy.Column('value_text', sqlalchemy.String),

            sqlalchemy.ForeignKeyConstraint(
                [
                    'dataprovider_id',
                    'train_session_id',
                    'train_id'
                ],
                [
                    self.xingu_db_table_prefix + 'training.dataprovider_id',
                    self.xingu_db_table_prefix + 'training.train_session_id',
                    self.xingu_db_table_prefix + 'training.train_id',
                ]
            ),
        )

        if 'awsathena' not in con['url']:
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_time',             self.tables[table_name].time),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_dataprovider_id',  self.tables[table_name].dataprovider_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_train_session_id', self.tables[table_name].train_session_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_train_id',         self.tables[table_name].train_id),


        table_name='metrics_estimation'
        self.tables[table_name] = sqlalchemy.Table(
            self.xingu_db_table_prefix + table_name,
            self.xingu_db_metadata,
            sqlalchemy.Column(
                'time', sqlalchemy.Integer,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_session_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'dataprovider_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'index', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'name', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column('value_number', sqlalchemy.Float),
            sqlalchemy.Column('value_text', sqlalchemy.String),

            sqlalchemy.ForeignKeyConstraint(
                [
                    'dataprovider_id',
                    'train_session_id',
                    'train_id'
                ],
                [
                    self.xingu_db_table_prefix + 'training.dataprovider_id',
                    self.xingu_db_table_prefix + 'training.train_session_id',
                    self.xingu_db_table_prefix + 'training.train_id',
                ]
            ),
        )

        if 'awsathena' not in con['url']:
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_time',             self.tables[table_name].time),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_dataprovider_id',  self.tables[table_name].dataprovider_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_train_session_id', self.tables[table_name].train_session_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_train_id',         self.tables[table_name].train_id),


        table_name='estimations'
        self.tables[table_name] = sqlalchemy.Table(
            self.xingu_db_table_prefix + table_name,
            self.xingu_db_metadata,
            sqlalchemy.Column(
                'time', sqlalchemy.Integer,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'dataprovider_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_session_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'train_id', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column(
                'index', sqlalchemy.String,
                primary_key=True,
                nullable=False
            ),
            sqlalchemy.Column('estimation', sqlalchemy.Float),

            sqlalchemy.ForeignKeyConstraint(
                [
                    'dataprovider_id',
                    'train_session_id',
                    'train_id'
                ],
                [
                    self.xingu_db_table_prefix + 'training.dataprovider_id',
                    self.xingu_db_table_prefix + 'training.train_session_id',
                    self.xingu_db_table_prefix + 'training.train_id',
                ]
            ),
        )

        if 'awsathena' not in con['url']:
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_time',             self.tables[table_name].time),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_dataprovider_id',  self.tables[table_name].dataprovider_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_train_session_id', self.tables[table_name].train_session_id),
            sqlalchemy.Index(self.xingu_db_table_prefix + table_name + '_by_train_id',         self.tables[table_name].train_id),


        self.logger.info('Going to create tables on Xingu DB')

        self.xingu_db_metadata.create_all()


import argparse
import json
import logging

from . import DataProvider
from . import DPSimple
from . import DataProviderFactory
from . import Coach
from . import Model
from . import ConfigManager



def prepare_logging(level=logging.INFO):
    # Switch between INFO/DEBUG while running in production/developping:

    # Configure logging for PanModel

    FORMATTER = logging.Formatter("%(asctime)s|%(levelname)s|%(name)s|%(message)s")
    HANDLER = logging.StreamHandler()
    HANDLER.setFormatter(FORMATTER)

    loggers=[
        logging.getLogger('__main__'),
        logging.getLogger('xingu'),
        logging.getLogger('sqlite')
    ]

    for logger in loggers:
        logger.addHandler(HANDLER)
        logger.setLevel(level)

    return loggers[0]



def prepare_args():
    parser = argparse.ArgumentParser(
        prog='xingu',
        description='Train and Batch Predict various Models for multiple DataProviders'
    )

    parser.add_argument(
        '--dps', '--data-providers',
        dest='DATAPROVIDER_LIST', required=False,
        default=ConfigManager().get('DATAPROVIDER_LIST', default=None),
        help=(
            'Comma separated list of DataProvider IDs to process. Overwrites '
            'DATAPROVIDER_LIST env. If not set, all classes inherited from '
            'DataProvider that have IDs will be used and a Model for each will '
            'be trained.'
        )
    )

    parser.add_argument(
        '--dps-folder', '--data-providers-folder',
        dest='DATAPROVIDER_FOLDER', required=False,
        default=ConfigManager().get('DATAPROVIDER_FOLDER', default=None),
        help=(
            'Folder name to scan for DataProviders. DataProviders are classes '
            'that hinerit from DataProvider and have an id set.'
        )
    )

    parser.add_argument(
        '--pre-req-train-or-session-ids',
        dest='PRE_REQ_TRAIN_OR_SESSION_IDS', required=False,
        default=ConfigManager().get('PRE_REQ_TRAIN_OR_SESSION_IDS', default=None),
        help=(
            'Comma-separated list to train or train session ID to search for '
            'pre-req models. Overwrites PRE_REQ_TRAIN_OR_SESSION_IDS env.'
        )
    )

    parser.add_argument(
        '--models-db',
        dest='XINGU_DB_URL', required=False,
        default=ConfigManager().get('XINGU_DB_URL',default='sqlite:///xingu.db?check_same_thread=False'),
        help=(
            'URL for Xingu´s control database as '
            '«mysql://user:pass@host.com/dbname?charset=utf8mb4». Overwrites '
            'XINGU_DB_URL env. If empty, uses '
            'sqlite:///xingu.db?check_same_thread=False, which is a SQLite '
            'database on current folder.'
        )
    )

    parser.add_argument(
        '--table-prefix', '--prefix',
        dest='XINGU_DB_TABLE_PREFIX',
        default=ConfigManager().get('XINGU_DB_TABLE_PREFIX',default=None),
        help=(
            'A string to prefix every Xingu DB table name with, such as '
            '“xingu_”. Overwrites XINGU_DB_TABLE_PREFIX env.'
        )
    )

    parser.add_argument(
        '--database', nargs="+",
        action="append", dest='DATABASES',
        default=ConfigManager().get('DATABASES',default=None),
        help=(
            'Takes 2 arguments: nickname and SQLAlchemy URL of a database. '
            'Can be used multiple times to define multiple databases. '
            'Overwrites DATABASES env.'
        )
    )

    parser.add_argument(
        '--hyperopt-strategy',
        dest='HYPEROPT_STRATEGY',
        default=ConfigManager().get('HYPEROPT_STRATEGY',default=None),
        help=(
            'Strategy for hyperparam optimization before training process. '
            'May be “last” or “self” or “dp” or a traind_id or a '
            'train_session_id or simply not set. Overwrites HYPEROPT_STRATEGY '
            'env. If None or not set, Estimator’s defaults will be used.'
        )
    )

    parser.add_argument(
        '--datasource-cache-path',
        dest='DATASOURCE_CACHE_PATH',
        default=ConfigManager().get('DATASOURCE_CACHE_PATH',default=None),
        help=(
            'Folder to store parquets of unprocessed DataProviders’ SQL queries '
            'results. Useful to speed up consecutive and repetitive runs in '
            'development scenarios. Overwrites DATASOURCE_CACHE_PATH env.'
        )
    )

    parser.add_argument(
        '--dvc-query-cache-path',
        dest='DVC_DATASOURCE_CACHE_PATH',
        default=ConfigManager().get('DVC_DATASOURCE_CACHE_PATH',default=None),
        help=(
            'Usually set this to the same as --datasource-cache-path. If set, '
            'causes DVC to commit DataProvider’s queries cache files. If set to '
            'a different folder, an additional (and unnecessary) copy of cache '
            'files will be created there, just for DVC. Overwrites '
            'DVC_DATASOURCE_CACHE_PATH env.'
        )
    )

    parser.add_argument(
        '--trained-models-path',
        dest='TRAINED_MODELS_PATH',
        default=ConfigManager().get('TRAINED_MODELS_PATH',default=None),
        help=(
            'A local folder or an S3 path to dump trained models’ pickles. '
            'Example: «s3://mlops-data/sample-dvc/teste-do-avi/avm-trained-models». '
            'Overwrites TRAINED_MODELS_PATH env.'
        )
    )

    parser.add_argument(
        '--plots-path',
        dest='PLOTS_PATH',
        default=ConfigManager().get('PLOTS_PATH',default=None),
        help=(
            'A local folder or an S3 path to save plots. Overwrites PLOTS_PATH '
            'env.'
        )
    )

    parser.add_argument(
        '--plots-format',
        dest='PLOTS_FORMAT',
        default=ConfigManager().get('PLOTS_FORMAT',default='png'),
        help=(
            'Comma separated list of image formats. Example: "svg,png". '
            'Defaults to PNG only. Overwrites PLOTS_FORMAT env.'
        )
    )

    parser.add_argument(
        '--dvc-trained-models-path',
        dest='DVC_TRAINED_MODELS_PATH',
        default=ConfigManager().get('DVC_TRAINED_MODELS_PATH',default=None),
        help=(
            'A local path to dump trained model pickles to later be added to '
            'DVC. Overwrites DVC_TRAINED_MODELS_PATH env.'
        )
    )

    parser.add_argument(
        '--train',
        dest='TRAIN', action=argparse.BooleanOptionalAction,
        default=ConfigManager().get('TRAIN',default=True, cast=bool),
        help='Set if you want to train models. Overwrites TRAIN env.'
    )

    parser.add_argument(
        '--train-session-id',
        dest='TRAIN_SESSION_ID',
        default=ConfigManager().get('TRAIN_SESSION_ID',default=None),
        help=(
            'Gives this training session an ID, otherwise randomname module '
            'will be used. Overwrites TRAIN_SESSION_ID env.'
        )
    )

    parser.add_argument(
        '--hostname',
        dest='HOSTNAME',
        default=ConfigManager().get('HOSTNAME',default=None),
        help=(
            'Sometimes the detected hostname is useless for reporting purposes. '
            'This is your chance to define something more useful. '
            'Overwrites HOSTNAME env.'
        )
    )

    parser.add_argument(
        '--username',
        dest='USERNAME',
        default=ConfigManager().get('USERNAME',default=None),
        help=(
            'Sometimes the detected user name is useless for reporting purposes. '
            'This is your chance to define something more useful. '
            'Overwrites USERNAME env.'
        )
    )

    parser.add_argument(
        '--parallel-train-max-workers',
        dest='PARALLEL_TRAIN_MAX_WORKERS',
        default=ConfigManager().get('PARALLEL_TRAIN_MAX_WORKERS', default=0),
        help=(
            'How many Models to train in parallel taking care of dependencies. '
            'Overwrites PARALLEL_TRAIN_MAX_WORKERS env.'
        )
    )

    parser.add_argument(
        '--parallel-hyperopt-max-workers',
        dest='PARALLEL_HYPEROPT_MAX_WORKERS',
        default=ConfigManager().get('PARALLEL_HYPEROPT_MAX_WORKERS', default=0),
        help=(
            'Control parallelism in hyper-parameter optimization. Overwrites '
            'PARALLEL_HYPEROPT_MAX_WORKERS env.'
        )
    )

    parser.add_argument(
        '--post-process',
        dest='POST_PROCESS', action=argparse.BooleanOptionalAction,
        default=ConfigManager().get('POST_PROCESS',default=True, cast=bool),
        help=(
            'Set if you want to post-process trains (save model, save data '
            'sets, batch predict, metrics). Usefull when hyper-parameters '
            'optimization is the goal. Overwrites POST_PROCESS env.'
        )
    )

    parser.add_argument(
        '--save-sets',
        dest='SAVE_SETS', action=argparse.BooleanOptionalAction,
        default=ConfigManager().get('SAVE_SETS',default=True, cast=bool),
        help=(
            'Wether to write training ‘sets’ to Xingu database table sets or not. '
            'Overwrites SAVE_SETS env.'
        )
    )

    parser.add_argument(
        '--parallel-post-process-max-workers',
        dest='PARALLEL_POST_PROCESS_MAX_WORKERS',
        default=ConfigManager().get('PARALLEL_POST_PROCESS_MAX_WORKERS', default=0),
        help=(
            'How many PanModels to post-process in parallel. Overwrites'
            'PARALLEL_POST_PROCESS_MAX_WORKERS env.'
        )
    )

    parser.add_argument(
        '--parallel-estimators-max-workers',
        dest='PARALLEL_ESTIMATORS_MAX_WORKERS',
        default=ConfigManager().get('PARALLEL_ESTIMATORS_MAX_WORKERS', default=0),
        help=(
            'How many estimators to train in parallel. Overwrites '
            'PARALLEL_ESTIMATORS_MAX_WORKERS env.'
        )
    )

    parser.add_argument(
        '--batch-predict',
        dest='BATCH_PREDICT', action=argparse.BooleanOptionalAction,
        default=ConfigManager().get('BATCH_PREDICT',default=True, cast=bool),
        help=(
            'Set if you want to batch predict data, which is required for many '
            'metrics computations. If --no-train, use pre-trained models from '
            'TRAINED_MODELS_PATH. Overwrites BATCH_PREDICT env.'
        )
    )

    parser.add_argument(
        '--batch-predict-save-estimations',
        dest='BATCH_PREDICT_SAVE_ESTIMATIONS',
        action=argparse.BooleanOptionalAction,
        default=ConfigManager().get('BATCH_PREDICT_SAVE_ESTIMATIONS',default=False, cast=bool),
        help=(
            'Set if you want to save to Xingu DB the batch predict '
            'estimations. Default is not to save. Batch predict metrics will '
            'always be saved. Overwrites BATCH_PREDICT_SAVE_ESTIMATIONS env.'
        )
    )

    parser.add_argument(
        '--dvc',
        dest='DVC', action=argparse.BooleanOptionalAction,
        default=ConfigManager().get('DVC',default=True, cast=bool),
        help=(
            'If --no-dvc, disables all DVC operations regardless of what other '
            'DVC-related variables have. Overwrites DVC env.'
        )
    )

    parser.add_argument(
        '--commit-inventory',
        dest='COMMIT_INVENTORY', action=argparse.BooleanOptionalAction,
        default=ConfigManager().get('COMMIT_INVENTORY',default=False, cast=bool),
        help=(
            'The inventory.yaml file will be updated at the end of train '
            'session. This flag controls weather it should be also commited to '
            'Git’s current branch and pushed to remote repo. Overwrites '
            'COMMIT_CURRENTS env.'
        )
    )

    parser.add_argument(
        '--project-home', dest='PROJECT_HOME',
        default=ConfigManager().get('PROJECT_HOME',default='.'),
        help=(
            'Local clone of xingu git repository, to collect various metadata '
            'during runtime. Overwrites PROJECT_HOME env.'
        )
    )



    # Simple DataProvider arguments (entire DataProvider constructed from
    # command line arguments, no coding)
    parser.add_argument(
        '--simpledp-id', dest='SIMPLEDP_ID',
        default=ConfigManager().get('SIMPLEDP_ID',default=None),
        help=('The ID of the Simple DataProvider. Overwrites SIMPLEDP_ID env.')
    )

    parser.add_argument(
        '--simpledp-train-datasource', nargs="+",
        action="append", dest='TRAIN_DATASOURCE',
        default=ConfigManager().get('TRAIN_DATASOURCE',default=None),
        help=(
            'Use multiple times. Pass URLs (http, S3, file etc) of CSVs, '
            'parquets or JSON files to define a training datasoure. Overwrites '
            'TRAIN_DATASOURCE env.'
        )
    )

    parser.add_argument(
        '--simpledp-estimator-features', dest='ESTIMATOR_FEATURES',
        default=ConfigManager().get('ESTIMATOR_FEATURES',default=None),
        help=(
            'Comma-separated list of column names for the estimator. '
            'Overwrites ESTIMATOR_FEATURES env.'
        )
    )

    parser.add_argument(
        '--simpledp-target-feature', dest='TARGET_FEATURE',
        default=ConfigManager().get('TARGET_FEATURE',default=None),
        help=(
            'Name of target column in the train datasource. Overwrites '
            'TARGET_FEATURE env.'
        )
    )

    parser.add_argument(
        '--simpledp-proba-class-index', dest='PROBA_CLASS_INDEX', type=int,
        default=ConfigManager().get('PROBA_CLASS_INDEX',default=0,cast=int),
        help=(
            'For classifiers, the index of the desired result of '
            'predict_proba(). Defaults to 0. Overwrites PROBA_CLASS_INDEX env.'
        )
    )

    parser.add_argument(
        '--simpledp-base-class', dest='BASE_CLASS',
        default=ConfigManager().get('BASE_CLASS',default=None),
        help=(
            'Python full name of a xingu.DataProvider-derived class to be used '
            'as a base class for the trained DataProvider. Overwrites '
            'BASE_CLASS env.'
        )
    )

    parser.add_argument(
        '--simpledp-estimator-class', dest='ESTIMATOR_CLASS',
        default=ConfigManager().get('ESTIMATOR_CLASS',default=None),
        help=(
            'Python full name of a xingu.Estimator-derived class, such as '
            '“xingu.estimators.xgboost_optuna.XinguXGBoostClassifier”. '
            'Overwrites ESTIMATOR_CLASS env.'
        )
    )

    parser.add_argument(
        '--simpledp-estimator-class-params', dest='ESTIMATOR_CLASS_PARAMS',
        default=ConfigManager().get('ESTIMATOR_CLASS_PARAMS',default=None),
        help=(
            'JSON text used as parameters to your xingu.Estimator-derived '
            'class. Overwrites ESTIMATOR_CLASS_PARAMS env.'
        )
    )

    parser.add_argument(
        '--simpledp-estimator-params', dest='ESTIMATOR_PARAMS',
        default=ConfigManager().get('ESTIMATOR_PARAMS',default=None),
        help=(
            'JSON text used as parameters for the underlying algorithm. '
            'Overwrites ESTIMATOR_PARAMS env.'
        )
    )

    parser.add_argument(
        '--simpledp-estimator-hyperparams', dest='ESTIMATOR_HYPERPARAMS',
        default=ConfigManager().get('ESTIMATOR_HYPERPARAMS',default=None),
        help=(
            'JSON text used as hyperparameters for the underlying algorithm. '
            'Together, estimator params and hyperparams are passed to the '
            'underlying algorithm, but hyperparams are usually subject to '
            'optimization. Overwrites ESTIMATOR_HYPERPARAMS env.'
        )
    )

    parser.add_argument(
        '--simpledp-estimator-hyperparams-search-space',
        dest='ESTIMATOR_HYPERPARAMS_SEARCH_SPACE',
        default=ConfigManager().get('ESTIMATOR_HYPERPARAMS_SEARCH_SPACE',default=None),
        help=(
            'JSON text of algorithm hyperparams range instrumented by the '
            'hyperparam optimizer. Overwrites '
            'ESTIMATOR_HYPERPARAMS_SEARCH_SPACE env.'
        )
    )



    parser.add_argument(
        '--debug', dest='DEBUG',
        action=argparse.BooleanOptionalAction,
        default=ConfigManager().get('DEBUG',default=False, cast=bool),
        help='Be more verbose and output messages to console.'
    )

    parsed = parser.parse_args()

    unset_if_none=[
        'DATAPROVIDER_LIST',             'XINGU_DB_TABLE_PREFIX',
        'QUERY_CACHE_PATH',              'DVC_QUERY_CACHE_PATH',
        'DVC_TRAINED_MODELS_PATH',       'PARALLEL_TRAIN_MAX_WORKERS',
        'HYPEROPT_STRATEGY',             'TRAINED_MODELS_PATH',
        'PRE_REQ_TRAIN_OR_SESSION_IDS',  'DATAPROVIDER_FOLDER',
        'HOSTNAME',                      'USERNAME',

        # Simple DataProvider configs
        'SIMPLEDP_ID',                   'BASE_CLASS',
        'TRAIN_DATASOURCE',              'TARGET_FEATURE',
        'ESTIMATOR_CLASS',               'ESTIMATOR_FEATURES',
        'PROBA_CLASS_INDEX',             'ESTIMATOR_CLASS_PARAMS',
        'ESTIMATOR_PARAMS',              'ESTIMATOR_HYPERPARAMS_SEARCH_SPACE',
        'ESTIMATOR_HYPERPARAMS'

    ]

    args={}
    for c in parsed.__dict__:
        if parsed.__dict__[c] is not None or c not in unset_if_none:
            args.update({c: parsed.__dict__[c]})

    return args



def main():
    # Read environment and command line parameters
    args=prepare_args()


    # Setup logging
    global logger
    if args['DEBUG']:
        logger=prepare_logging(logging.DEBUG)
    else:
        logger=prepare_logging()

    simpledp = None
    if 'SIMPLEDP_ID' in args:
        simpledp = DPSimple(**args)


    # Gather all DataProviders requested
    if 'DATAPROVIDER_LIST' in args:
        dpf = DataProviderFactory(
            providers_list=[x.strip() for x in args['DATAPROVIDER_LIST'].split(',')],
            providers_folder=args['DATAPROVIDER_FOLDER'],
            providers_extra_objects=simpledp,
        )
    else:
        dpf = DataProviderFactory(
            providers_folder=args['DATAPROVIDER_FOLDER'],
            providers_extra_objects=simpledp,
        )

    # Prepare list of databases to be parsed latter by Coach.get_db_connection()
    if isinstance(args['DATABASES'],list):
        args['DATABASES'] = '|'.join(args['DATABASES'])

    # Control DVC
    if args['DVC'] == False:
        dvc_killer=['DVC_QUERY_CACHE_PATH', 'DVC_TRAINED_MODELS_PATH']
        for dvc in dvc_killer:
            if dvc in args: del args[dvc]


    # Now set the environment with everything useful that we got
    ConfigManager().set(args)


    # Setup a coach
    coach=Coach(dp_factory = dpf)


    if args['TRAIN']:
        coach.team_train()
    elif args['BATCH_PREDICT']:
        coach.team_batch_predict()
    else:
        logger.info('Please use both or one of --batch-predict and --train')





if __name__ == "__main__":
    main()



import sys
import pathlib
import pkgutil
import importlib
import inspect
import logging

from . import DataProvider
from . import ConfigManager





class DataProviderFactory:

    def __init__(self, providers_folder: str=None, providers_list: list=None, providers_extra_objects=None):
        # 1. Check if providers_folder is valid
        # 2. Load all DPs whose IDs are in providers_list. Load all if None
        # 3. If no providers_folder passed, search for config_file
        # 4. Look for providers_folder in config file
        # 5. Look for providers_list in config file
        # 6. Check for providers in xingu.dataproviders


        # Setup logging
        self.logger = logging.getLogger(__name__ + '.' + self.__class__.__name__)


        # Initialize parameters to something useless and fill them up along the way
        self.providers_folder   = None
        self.providers_list     = None
        self.config             = None
        self.dps                = {}


        # Now overwrite whatever came from config file with what was passed to __init__
        if providers_folder is not None:
            self.providers_folder = providers_folder
        else:
            self.providers_folder = ConfigManager().get('DATAPROVIDER_FOLDER', default=None)

        if providers_list is not None:
            self.providers_list = providers_list
        else:
            self.providers_list = ConfigManager().get('DATAPROVIDER_LIST', default=None)

        if isinstance(self.providers_list, list):
            # Remove useless elements
            self.providers_list = list(filter(None, self.providers_list))
            if (len(self.providers_list) < 1):
                self.providers_list = None
        elif self.providers_list is not None:
            self.providers_list = [i.strip() for i in self.providers_list.split(',')]


        # If we still don't have a path to work with, use current folder
        if self.providers_folder is None:
            self.providers_folder = str(pathlib.Path('.'))

        self.logger.info("Folder to search DataProviders: {}".format(self.providers_folder))

        # We are now configured as much as we could.

        # Insert self.providers_folder in modules path in order to
        # make importlib.import_module() work properly
        sys.path.insert(1, self.providers_folder)


        # Load and store classes which IDs on self.providers_list
        for modinfo in pkgutil.iter_modules([self.providers_folder]):
            # Import the python file as a module
            mod=importlib.import_module(modinfo.name)
            self.logger.debug("Imported module {}".format(modinfo.name))

            # Walk through each thing inside that module and check if it is a desired DP
            for thing_name in dir(mod):
                thing = getattr(mod, thing_name)

                if (
                        # Check if class
                        inspect.isclass(thing)

                        # Check if subclass of DataProvider
                        and issubclass(thing, DataProvider)

                        # Check if we have and ID
                        and hasattr(thing, 'id')

                        # Check if ID is meaningful
                        and thing.id is not None

                        # Check if not DataProvider itself
                        and thing != DataProvider

#                         and (
#                             # Check if we have a list of desired classes, use it anyway otherwise
#                             self.providers_list is None or
#
#                             # Check if class is desired through its ID
#                             thing.id in self.providers_list
#                         )
                ):
                    self.dps.update({thing.id: thing})

        # Process DataProvider objects passed on providers_extra_objects
        if providers_extra_objects is not None:
            if issubclass(type(providers_extra_objects),DataProvider):
                providers_extra_objects=[providers_extra_objects]
            if isinstance(providers_extra_objects,list):
                self.dps.update({
                    dp.id: dp
                    for dp in providers_extra_objects
                })

        if self.providers_list is None:
            self.providers_list=list(self.dps.keys())



    def produce(self):
        for dp in self.providers_list:
            # Return a concrete class instance of a true real DataProvider implementation
            try:
                if type(self.dps[dp])==type:
                    yield self.dps[dp]()
                else:
                    yield self.dps[dp]
            except KeyError as e:
                self.logger.exception(f'Can’t find DataProvider with ID «{dp}».')



    def get_pre_req(self, dp):
        """
        Return a set of IDs of all pre-req models for dp.
        So you'll get {cartorios,anuncios} for dp=vendas_loft.
        """
        pre_req=set()
        sub_pre_req=set()

        if isinstance(dp, str):
            # Convert a DataProvider ID into a real DataProvider
            dp=self.dps[dp]

        if hasattr(dp, 'pre_req'):
            pre_req=pre_req | dp.pre_req

            if len(pre_req):
                for p in pre_req:
                    sub_pre_req=sub_pre_req | self.get_pre_req(p)

        return pre_req | sub_pre_req



    def __repr__(self):
        return 'DataProviderFactory(' + self.dps.__repr__() + ')'